#ifndef CLASSIFIERAPP_H
#define CLASSIFIERAPP_H 

#include <string>
#include <fstream>
#include <functional>

#include "tet/app/appbase.h"

#include "tet/descriptors/descriptor.h"
#include "tet/encodings/encoding.h"

#include "tet/containers/view.h"
#include "tet/containers/instance.h"
#include "tet/containers/class.h"
#include "tet/containers/set.h"

#include "tet/classifiers/svm.h"

#include "tet/processing/dataprocessing.h"

#include "tet/io/xmlserializable.h"

namespace tet
{

namespace classification
{

namespace app
{

template <class _DescriptorType, class _EncoderType>
class ClassifierApp : public AppBase, public tet::classification::io::XMLSerializable
{
public:
    typedef _DescriptorType DescriptorType;
    typedef _EncoderType EncoderType;

    typedef typename DescriptorType::InputType InputType;
    typedef typename DescriptorType::FeatureType FeatureType;
    typedef typename EncoderType::FeatureType EncodingType;

    typedef tet::classification::containers::View     View;
    typedef tet::classification::containers::Instance Instance;
    typedef tet::classification::containers::Class    Class;
    typedef tet::classification::containers::Set      Set;

    ClassifierApp() : _useCached(false)
                    , _useCustom(false)
                    , _name("default")
                    , _cacheDir(".")
                    , _getSetDescriptors(_cacheDir)
                    , _getSetEncodings(_cacheDir) {}

    virtual const char * xmlName() const { return "classifierApp"; }

    void setDescriptor(DescriptorType* descriptor)
    { _descriptor = descriptor; }
    void setEncoding(EncoderType* encoding)
    { _encoding = encoding; }

    void setUseCached(bool useCached) { _useCached = useCached; }
    void setUseCustom(bool useCustom) { _useCustom = useCustom; }

    void setKernel(kernels::KernelBase::Ptr kernel)
    { _kernel = kernel; _svm.setKernel(kernel); }

    const DescriptorType & getDescriptor() { return _descriptor; }
    const EncoderType & getEncoding() { return _encoding; }

    const Set & getPositives() { return _positives; }
    const Set & getNegatives() { return _negatives; }
    const Set & getTests()     { return _tests; }
    const Set & getCustom()    { return _custom; }

    void run();
    void train();
    void test();
    void testCoeff(std::string weightsFile);

    void setSVMModelFile(std::string modelFile)
    { _svm.setModelFile(modelFile); }

    const tet::classification::classifiers::SVM & getSVM()
    { return _svm; }

    // serialization
    void readFile(std::string filename);
    void writeFile(std::string filename);
    tinyxml2::XMLElement* serialize(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc);
    void deserialize(const tinyxml2::XMLElement*);

protected:

private:
    bool _useCached;
    bool _useCustom;

    kernels::KernelBase::Ptr _kernel;

    std::string _name;
    std::string _cacheDir;

    processing::GetSetDescriptorsCached<DescriptorType> _getSetDescriptors;
    processing::GetSetEncodingsCached<DescriptorType, EncoderType> _getSetEncodings;

    Set _positives, _negatives, _tests, _custom;

    DescriptorType _descriptor;
    EncoderType _encoding;

    std::string _resultFile;

    tet::classification::classifiers::SVM _svm;

}; // class ClassifierApp

} // namespace app

} // namespace classification

} // namespace tet

#include "impl/classifierapp.hpp"

#endif /* CLASSIFIERAPP_H */
