#include <fstream>

#include "tet/app/classifierapp.h"
#include "tet/util/printutils.h"

#include "tet/io/sampleloader.h"
#include "tet/io/svmlight.h"

#include "tet/classifiers/svm.h"

#include "tet/util/samplingutils.h"

#include "tet/processing/dataprocessing.h"

namespace tet
{

namespace classification
{

namespace app
{

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::run()
{
    TETLOG("Start..."); 
    train();
    test();
    TETLOG("Done."); 
}

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::train()
{
    std::stringstream weightFilenameStr;
    weightFilenameStr << "weights_" << _name << ".txt";

    std::vector<Eigen::VectorXf*> posEncodings = _getSetEncodings(_positives, _descriptor, _encoding);
    std::vector<Eigen::VectorXf*> negEncodings = _getSetEncodings(_negatives, _descriptor, _encoding);
    _svm.train_classifier(posEncodings, negEncodings, weightFilenameStr.str());

    if (_kernel->isLinear())
    {
        tet::classification::io::updateEncoderWithWeights<_DescriptorType, _EncoderType>(&_encoding, weightFilenameStr.str());
    }

}

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::test()
{
    std::vector<Eigen::VectorXf*> testEncodings = _getSetEncodings(_tests, _descriptor, _encoding);
    _svm.test_classifier(testEncodings, ".svmTestResults.txt");

    std::vector<Eigen::VectorXf*> posEncodings = _getSetEncodings(_positives, _descriptor, _encoding);
    _svm.test_classifier(posEncodings, ".svmPositivesResults.txt");

    std::vector<Eigen::VectorXf*> negEncodings = _getSetEncodings(_negatives, _descriptor, _encoding);
    _svm.test_classifier(negEncodings, ".svmNegativesResults.txt");

    tet::classification::io::updateSetWithScores(_tests, ".svmTestResults.txt");
    tet::classification::io::updateSetWithScores(_positives, ".svmPositivesResults.txt");
    tet::classification::io::updateSetWithScores(_negatives, ".svmNegativesResults.txt");
}

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::testCoeff(std::string weightsFile)
{
    std::vector<Eigen::VectorXf*> testEncodings = _getSetEncodings(_tests, _descriptor, _encoding);
    _svm.test_classifier_existing(testEncodings, ".svmTestResults.txt", weightsFile);

    std::vector<Eigen::VectorXf*> posEncodings = _getSetEncodings(_positives, _descriptor, _encoding);
    _svm.test_classifier_existing(posEncodings, ".svmPositivesResults.txt", weightsFile);

    std::vector<Eigen::VectorXf*> negEncodings = _getSetEncodings(_negatives, _descriptor, _encoding);
    _svm.test_classifier_existing(negEncodings, ".svmNegativesResults.txt", weightsFile);

    tet::classification::io::updateSetWithScores(_tests, ".svmTestResults.txt");
    tet::classification::io::updateSetWithScores(_positives, ".svmPositivesResults.txt");
    tet::classification::io::updateSetWithScores(_negatives, ".svmNegativesResults.txt");
}

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::readFile(std::string filename)
{
    tinyxml2::XMLDocument doc;
    if (doc.LoadFile(filename.c_str()) != tinyxml2::XML_SUCCESS)
    {
        funcMsg("ClassifierApp::readFile", "Reading file", "Error: Can't read file");
    }

    if (const tinyxml2::XMLElement *root = doc.FirstChildElement("file"))
    {
        deserialize(root->FirstChildElement("settings"));
    }
}

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::writeFile(std::string filename)
{
    tinyxml2::XMLDocument doc;
    tinyxml2::XMLDeclaration* decl = doc.NewDeclaration("xml version=\"1.0\"");
    doc.LinkEndChild(decl);

    tinyxml2::XMLComment* comment = doc.NewComment(" Settings for TET Classification Suite (classification)");
    doc.LinkEndChild(comment);

    tinyxml2::XMLElement* root = doc.NewElement("file");
    serialize(root, &doc);
    doc.LinkEndChild(root);

    doc.SaveFile(filename.c_str());

}

template <class _DescriptorType, class _EncoderType>
tinyxml2::XMLElement* ClassifierApp<_DescriptorType, _EncoderType>::serialize(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *root = doc->NewElement("settings");
    parent->LinkEndChild(root);

    tinyxml2::XMLElement *cacheDir = doc->NewElement("cacheDir");
    cacheDir->SetText(_cacheDir.c_str());
    root->LinkEndChild(cacheDir);

    _kernel->serialize(root, doc);

    tinyxml2::XMLElement *positives = doc->NewElement("positives");
    _positives.serialize(positives, doc);
    root->LinkEndChild(positives);

    tinyxml2::XMLElement *negatives = doc->NewElement("negatives");
    _negatives.serialize(negatives, doc);
    root->LinkEndChild(negatives);

    tinyxml2::XMLElement *tests = doc->NewElement("tests");
    _tests.serialize(tests, doc);
    root->LinkEndChild(tests);

    tinyxml2::XMLElement *setName = doc->NewElement("name");
    setName->SetText(_name.c_str());
    root->LinkEndChild(setName);

    _descriptor.serialize(root, doc);
    _encoding.serialize(root, doc);

    return root;
}

template <class _DescriptorType, class _EncoderType>
void ClassifierApp<_DescriptorType, _EncoderType>::deserialize(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *cacheDir = root->FirstChildElement("cacheDir"))
    {
        _cacheDir = std::string(cacheDir->GetText());
        _getSetDescriptors = processing::GetSetDescriptorsCached<_DescriptorType>(std::string(cacheDir->GetText()));
        _getSetEncodings   = processing::GetSetEncodingsCached<_DescriptorType, _EncoderType>(std::string(cacheDir->GetText()));
    }

    if (const tinyxml2::XMLElement *kernelTag = root->FirstChildElement("kernel"))
    {
        std::string type = std::string(kernelTag->Attribute("type"));
        kernels::KernelBase::Ptr kernel = kernels::KernelFactory::get(type);
        kernel->deserialize(kernelTag);
        setKernel(kernel);
    }

    if (const tinyxml2::XMLElement *descriptor = root->FirstChildElement("descriptor"))
    {
        _descriptor.deserialize(descriptor);
    }

    if (const tinyxml2::XMLElement *encoding = root->FirstChildElement("encoding"))
    {
        _encoding.deserialize(encoding);
        _encoding.init(_descriptor, _getSetDescriptors);
    }

    if (const tinyxml2::XMLElement *setName = root->FirstChildElement("name"))
    {
        _name = std::string(setName->GetText());
    }

    if (const tinyxml2::XMLElement *positives = root->FirstChildElement("positives"))
    {
        if (const tinyxml2::XMLElement *posSet = positives->FirstChildElement("set"))
        {
            _positives.deserialize(posSet);
        } else if (const tinyxml2::XMLElement *posFile = positives->FirstChildElement("file"))
        {
            _positives.deserializeFromFile(std::string(posFile->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *negatives = root->FirstChildElement("negatives"))
    {
        if (const tinyxml2::XMLElement *negSet = negatives->FirstChildElement("set"))
        {
            _negatives.deserialize(negSet);
        } else if (const tinyxml2::XMLElement *negFile = negatives->FirstChildElement("file"))
        {
            _negatives.deserializeFromFile(std::string(negFile->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *tests = root->FirstChildElement("tests"))
    {
        if (const tinyxml2::XMLElement *testSet = tests->FirstChildElement("set"))
        {
            _tests.deserialize(testSet);
        } else if (const tinyxml2::XMLElement *testFile = tests->FirstChildElement("file"))
        {
            _tests.deserializeFromFile(std::string(testFile->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *custom = root->FirstChildElement("custom"))
    {
        if (const tinyxml2::XMLElement *customSet = custom->FirstChildElement("set"))
        {
            _custom.deserialize(customSet);
        } else if (const tinyxml2::XMLElement *customFile = custom->FirstChildElement("file"))
        {
            _custom.deserializeFromFile(std::string(customFile->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *useCached = root->FirstChildElement("useCached"))
    {
        setUseCached(std::string(useCached->GetText()) == "1" ? true : false);
    }


}

} // namespace app

} // namespace classification

} // namespace tet
