#include <fstream>

#include "tet/app/regressorapp.h"

#include "tet/util/printutils.h"

#include "tet/io/sampleloader.h"
#include "tet/io/svmlight.h"

#include "tet/classifiers/svm.h"

#include "tet/eval/evalmetricfactory.h"

namespace tet
{

namespace classification
{

namespace app
{

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::printInfo() const
{
    std::cout << "Regressor name  : " << _name << std::endl;
    std::cout << "Cache dir       : " << _cacheDir << std::endl;
    std::cout << "Kernel type     : " << _kernel->toStr() << std::endl;
}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::run()
{
    TETLOG("Starting regressor...");
    printInfo();
    train();
    test();
    TETLOG("Regressor done.");
}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::train()
{
    std::stringstream weightFilenameStr;
    weightFilenameStr << "weights_" << _name << ".txt";

    std::vector<Eigen::VectorXf*> sampleEncodings = _getSetEncodings(_samples, _descriptor, _encoder);
    std::vector<float> sampleGT = tet::classification::containers::getGTs(_samples);

    _svm.train_regressor(sampleEncodings, sampleGT, weightFilenameStr.str());

    if (_kernel->isLinear())
    {
        tet::classification::io::updateEncoderWithWeights<_DescriptorType, _EncoderType>(&_encoder, weightFilenameStr.str());
    }

}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::test()
{
    std::vector<Eigen::VectorXf*> testEncodings = _getSetEncodings(_tests, _descriptor, _encoder);
    _svm.test_regressor(testEncodings, ".svmTestResults.txt");
    tet::classification::io::updateSetWithScores(_tests, ".svmTestResults.txt");

    std::vector<Eigen::VectorXf*> sampleEncodings = _getSetEncodings(_samples, _descriptor, _encoder);
    _svm.test_regressor(sampleEncodings, ".svmSampleResults.txt");
    tet::classification::io::updateSetWithScores(_samples, ".svmSampleResults.txt");

    tet::classification::eval::EvalData setEvalData =
        tet::classification::containers::getEvalData(_tests);

    _eval.computeMetrics(setEvalData);

}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::setUseCached
(bool useCached)
{
    _useCached = useCached;
}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::readFile(std::string filename)
{
    tinyxml2::XMLDocument doc;
    if (doc.LoadFile(filename.c_str()) != tinyxml2::XML_SUCCESS)
    {
        TETLOG("Error: Can't read file");
    }

    if (const tinyxml2::XMLElement *root = doc.FirstChildElement("file"))
    {
        deserialize(root->FirstChildElement("settings"));
    }
}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::writeFile(std::string filename)
{
    tinyxml2::XMLDocument doc;
    tinyxml2::XMLDeclaration* decl = doc.NewDeclaration("xml version=\"1.0\"");
    doc.LinkEndChild(decl);

    tinyxml2::XMLComment* comment = doc.NewComment(" Settings for TET Classification Suite (regression)");
    doc.LinkEndChild(comment);

    tinyxml2::XMLElement* root = doc.NewElement("file");
    serialize(root, &doc);
    doc.LinkEndChild(root);

    doc.SaveFile(filename.c_str());

}

template <class _DescriptorType, class _EncoderType>
tinyxml2::XMLElement* RegressorApp<_DescriptorType, _EncoderType>::serialize(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *root = doc->NewElement("settings");
    parent->LinkEndChild(root);

    tinyxml2::XMLElement *cacheDir = doc->NewElement("cacheDir");
    cacheDir->SetText(_cacheDir.c_str());
    root->LinkEndChild(cacheDir);

    _kernel->serialize(root, doc);

    tinyxml2::XMLElement *eps = doc->NewElement("eps");
    eps->SetText(_eps);
    root->LinkEndChild(eps);

    tinyxml2::XMLElement *c = doc->NewElement("c");
    c->SetText(_c);
    root->LinkEndChild(c);

    tinyxml2::XMLElement *samples = doc->NewElement("samples");
    _samples.serialize(samples, doc);
    root->LinkEndChild(samples);

    tinyxml2::XMLElement *tests = doc->NewElement("tests");
    _tests.serialize(tests, doc);
    root->LinkEndChild(tests);

    tinyxml2::XMLElement *setName = doc->NewElement("name");
    setName->SetText(_name.c_str());
    root->LinkEndChild(setName);

    _descriptor.serialize(root, doc);
    _encoder.serialize(root, doc);

    for (unsigned int evalIdx = 0;
         evalIdx < _eval.size();
         ++evalIdx)
    {
        _eval.get(evalIdx)->serialize(root, doc);
    }

    return root;
}

template <class _DescriptorType, class _EncoderType>
void RegressorApp<_DescriptorType, _EncoderType>::deserialize(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *cacheDir = root->FirstChildElement("cacheDir"))
    {
        _cacheDir = std::string(cacheDir->GetText());
        _getSetDescriptors = processing::GetSetDescriptorsCached<_DescriptorType>(std::string(cacheDir->GetText()));
        _getSetEncodings   = processing::GetSetEncodingsCached<_DescriptorType, _EncoderType>(std::string(cacheDir->GetText()));
    }

    if (const tinyxml2::XMLElement *kernelTag = root->FirstChildElement("kernel"))
    {
        std::string type = std::string(kernelTag->Attribute("type"));
        kernels::KernelBase::Ptr kernel = kernels::KernelFactory::get(type);
        kernel->deserialize(kernelTag);
        setKernel(kernel);
    }

    if (const tinyxml2::XMLElement *eps = root->FirstChildElement("eps"))
    {
        eps->QueryFloatText(&_eps);
        _svm.setEps(_eps);
    }

    if (const tinyxml2::XMLElement *c = root->FirstChildElement("c"))
    {
        c->QueryFloatText(&_c);
        _svm.setC(_c);
    }

    if (const tinyxml2::XMLElement *descriptor = root->FirstChildElement("descriptor"))
    {
        _descriptor.deserialize(descriptor);
    }

    if (const tinyxml2::XMLElement *setName = root->FirstChildElement("name"))
    {
        _name = std::string(setName->GetText());
    }

    if (const tinyxml2::XMLElement *samples = root->FirstChildElement("samples"))
    {
        if (const tinyxml2::XMLElement *sampleSet = samples->FirstChildElement("set"))
        {
            _samples.deserialize(sampleSet);
        } else if (const tinyxml2::XMLElement *sampleXMLFile = samples->FirstChildElement("file"))
        {
            _samples.deserializeFromFile(std::string(sampleXMLFile->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *encoding = root->FirstChildElement("encoding"))
    {
        _encoder.deserialize(encoding);
        _encoder.init(_descriptor, _getSetDescriptors);
    }


    if (const tinyxml2::XMLElement *tests = root->FirstChildElement("tests"))
    {
        if (const tinyxml2::XMLElement *testSet = tests->FirstChildElement("set"))
        {
            _tests.deserialize(testSet);
        } else if (const tinyxml2::XMLElement *testsXMLFile = tests->FirstChildElement("file"))
        {
            _tests.deserializeFromFile(std::string(testsXMLFile->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *useCached = root->FirstChildElement("useCached"))
    {
        setUseCached(std::string(useCached->GetText()) == "1" ? true : false);
    }

    for (const tinyxml2::XMLElement *evalTag = root->FirstChildElement("eval");
         evalTag != NULL;
         evalTag = evalTag->NextSiblingElement("eval"))
    {
        std::string type = std::string(evalTag->Attribute("type"));
        eval::EvalMetric::Ptr eval = eval::EvalMetricFactory::get(type);
        eval->deserialize(evalTag);

        std::cout << "Adding metric " << type << std::endl;
        _eval.addMetric(eval);
    }

}

} // namespace app

} // namespace classification

} // namespace tet
