#ifndef HOGDESCRIPTOR_H
#define HOGDESCRIPTOR_H

#include <vector>

#include <boost/functional/hash.hpp>

#include "eigen3/Eigen/Core"

#include "tet/descriptors/descriptor.h"
#include "tet/io/xmlserializable.h"

#include "opencv2/opencv.hpp"

#include "vl/hog.h"

namespace tet
{
    namespace classification
    {
        namespace descriptors
        {
            class HOGDescriptor;
        }
    }
}

namespace std
{

    template <> struct hash<tet::classification::descriptors::HOGDescriptor>
    {
        size_t operator()(const tet::classification::descriptors::HOGDescriptor & desc) const;
    };

} // namespace std

namespace tet
{

namespace classification
{

namespace descriptors
{

    namespace _internal
    {
        struct HOGDescriptorTraits
        {
            typedef cv::Mat InputType;
            typedef Eigen::MatrixXf FeatureType;
            typedef cv::Mat VisType;

            typedef float MatType;
        };
    }

class HOGDescriptor : public Descriptor<_internal::HOGDescriptorTraits>
{
public:
    HOGDescriptor() : _nbins(9), _cellSize(8), 
                      _removeZeroGradient(true), _includeLocation(false)
    { 
        _vlHog = vl_hog_new(VlHogVariantDalalTriggs, 9, VL_FALSE);
    }

    Descriptor<_internal::HOGDescriptorTraits>* clone() const
    {
        HOGDescriptor* desc = new HOGDescriptor;

        desc->_nbins = _nbins;
        desc->_cellSize = _cellSize;
        desc->_removeZeroGradient = _removeZeroGradient;
        desc->_includeLocation = _includeLocation;

        return desc;
    }

    void setBins(int bins) { _nbins = bins; }
    int getBins() const { return _nbins; }

    void setCellSize(int cellSize) { _cellSize = cellSize; }
    int getCellSize() const { return _cellSize; }

    FeatureType getDescriptors(const InputType &img, VisType* vis=nullptr) const;
    VisType     getVisualization(const FeatureType &features) const;

    VlHog* getHog()
    { return _vlHog; }

    void setRemoveZeroGradient(bool removeZeroGradient)
    { _removeZeroGradient = removeZeroGradient; }
    
    ~HOGDescriptor()
    { vl_hog_delete(_vlHog); }
    
    // serialization
    tinyxml2::XMLElement* serialize(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc);
    void deserialize(const tinyxml2::XMLElement*);

    // hashing
    friend size_t std::hash<HOGDescriptor>::operator()(const HOGDescriptor&) const;
    friend std::size_t hash_value(const HOGDescriptor &);

protected:

private:
    int _nbins, _cellSize;
    bool _removeZeroGradient;
    bool _includeLocation;

    VlHog* _vlHog;
}; // class HOGDescriptor

std::size_t hash_value(const HOGDescriptor & desc);

} // namespace descriptors

} // namespace classification

} // namespace tet

#endif /* HOGDESCRIPTOR_H */
