#ifndef ENCODING_H
#define ENCODING_H 

#include <vector>
#include <functional>

#include "tet/processing/dataprocessing.h"
#include "tet/io/xmlserializable.h"

namespace tet
{

namespace classification
{

namespace encodings
{

//! Virtual class to be inherited by all encodings.
/*!
  This class is meant solely as a superclass to all feature encoding classes.
*/

template <class _EncodingTraits>
class Encoding : public tet::classification::io::XMLSerializable
{
public:
    typedef typename _EncodingTraits::Scalar      Scalar;
    typedef typename _EncodingTraits::FeatureType FeatureType;
    typedef typename _EncodingTraits::SampleType  SampleType;

    Encoding() : _computed(false), _usePCA(false), _pcaDim(-1) {}

    virtual const char * xmlName() const { return "encoding"; }

    virtual FeatureType computeFeatures(SampleType samples) const = 0;

    const std::vector<Scalar> & getFeatureWeights() const
    { return _featureWeights; }

    void setFeatureWeights(const std::vector<Scalar> & weights)
    { _featureWeights = weights; }

    //! NB: Derived classes *must* specify this template
    template <class _DescriptorType> void init
    (const _DescriptorType &, tet::classification::processing::GetSetDescriptorsBase<_DescriptorType> & descGetter) = delete;
    
    //! Returns if clusters have been computed
    bool isComputed() { return _computed; }

    //! Returns if encoding uses PCA
    bool usesPCA() { return _usePCA; }
    int getPCADim() { return _pcaDim; }
    
    // serialization
    tinyxml2::XMLElement* serialize(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc);
    void deserialize(const tinyxml2::XMLElement*);

    // to be specialized by all encodings
    virtual tinyxml2::XMLElement* serializeSpec(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc) = 0;
    virtual void deserializeSpec(const tinyxml2::XMLElement*) = 0;

protected:
    bool _computed;
    bool _useCached;
    bool _usePCA;
    int _pcaDim;

private:
    std::vector<Scalar> _featureWeights;

    size_t _hash;

}; // class Encoding

} // namespace encodings

} // namespace classification

} // namespace tet

#include "impl/encoding.hpp"

#endif /* ENCODING_H */
