#ifndef HISTENCODING_H
#define HISTENCODING_H 

#include "eigen3/Eigen/Core"
#include "opencv2/opencv.hpp"

#include "tet/encodings/encoding.h"
#include "tet/util/anntree.h"
#include "tet/io/xmlserializable.h"

#include <functional>

/*namespace std
{

    template <> struct hash<tet::classification::encodings::HistEncoding>
    {
        size_t operator()(const tet::classification::encodings::HistEncoding & enc) const;
    }

} // namespace std
*/

namespace tet
{

namespace classification
{

namespace encodings
{

    namespace _internal
    {
        struct HistEncodingTraits {
            typedef float                         Scalar;
            typedef Eigen::VectorXf               FeatureType;
            typedef Eigen::Matrix<Scalar, -1, -1> SampleType;

            struct HistEncodingInitStruct {
                SampleType* samples;
            };

            typedef HistEncodingInitStruct InitStruct;

            static constexpr const char* const name = "histogram";
        };


    }

//! Histogram Encoding class.
/*!
  This class implements Histogram Encoding of features using K-Means clustering.
  See http://www.robots.ox.ac.uk/~vgg/research/encoding_eval/
*/

class HistEncoding : public Encoding<_internal::HistEncodingTraits>
{
public:
    //! Constructor.
    /*!
      \param k number of bins (and thus features) in encoding
    */
    HistEncoding(int k=256) : _k(k), _usePCA(false) {}

    //! Clone.
    /*!
      Clones this histogram encoding, returning a pointer to the new one.
    */
    /*Encoding<_internal::HistEncodingTraits>* clone()
    {
        HistEncoding* enc = new HistEncoding();
        enc->_k = _k;
        enc->_usePCA = _usePCA;
        enc->_pcaDim = _pcaDim;
        enc->_pcaBasis = _pcaBasis;

        enc->_centers = _centers;
        enc->_tree = new tet::classification::util::ANNTree<Scalar>(*_tree);

        return enc;
    }*/

    //! Sets number of bins
    /*!
      Sets the number of bins for the encoding.
      \param k number of bins (and thus features) in encoding
    */
    void setK(int k) { _k = k; _computed = false; }
    int getK() const { return _k; }

    //! Sets number of PCA dimensions
    /*!
      Sets the number of PCA dimensions to use.
      \param dim number of dimensions to use
    */
    void setPCA(int dim)
    {
        if (dim < 0) { _usePCA = false; } else { _usePCA = true; }
        _pcaDim = dim;
        _computed = false;
    }

    //! Initialize encoding from samples
    /*!
      Computes cluster centers for histogram encoding.
      \param samples input samples, one row per sample
    */
    template <class _DescriptorType>
    void init(const _DescriptorType & descriptor, 
        tet::classification::processing::GetSetDescriptorsBase<_DescriptorType> & descGetter);

    //! Generate histogram encoding from samples
    /*!
      Generates histogram encoding from the samples given. If clusters
      were computed before, the encoding is generated based on these clusters.
      \param samples input samples, one row per sample
      \param recompute if true, clusters are recomputed even if clusters were computed before
      \return histogram of cluster assignments
    */
    FeatureType computeFeatures(SampleType samples) const;

    //! Returns computed cluster centers
    SampleType getCenters() const { return _centers; }
    void printCenter(size_t);
    void saveCenters();

    //! Returns closest center
    size_t getClosestCenter(SampleType sample) const
    { return _tree->findNN(sample.row(0)); }

    // serialization
    tinyxml2::XMLElement* serializeSpec(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc);
    void deserializeSpec(const tinyxml2::XMLElement*);

protected:

private:
    int _k;

    bool _usePCA;
    int _pcaDim;
    SampleType _pcaBasis; // computed PCA basis, if applicable

    SampleType _centers; // computed cluster centers
    tet::classification::util::ANNTree<Scalar> *_tree; // tree for quick neighbour lookup

}; // class HistEncoding

} // namespace encodings

} // namespace classification

} // namespace tet

/*namespace std
{

    template <> struct hash<tet::classification::encodings::HistEncoding>
    {
        size_t operator()(const tet::classification::encodings::HistEncoding & enc) const
        {
            size_t k = std::hash<int>(enc._k);
            size_t centers = 
        }
    }

} // namespace std
*/

#endif /* HISTENCODING_H */
