#ifndef KCENCODING_H
#define KCENCODING_H

#include "eigen3/Eigen/Core"
#include "opencv2/opencv.hpp"

#include "tet/containers/set.h"

#include "tet/encodings/encoding.h"

#include "tet/util/anntree.h"
#include "tet/util/eigenutils.h"

#include "tet/io/xmlserializable.h"

namespace tet
{

namespace classification
{

namespace encodings
{

    namespace _internal
    {
        struct KCEncodingTraits {
            typedef float                         Scalar;
            typedef Eigen::VectorXf               FeatureType;
            typedef Eigen::Matrix<Scalar, -1, -1> SampleType;

            struct KCEncodingInitStruct {
                SampleType* samples;
            };

            typedef KCEncodingInitStruct InitStruct;

            static constexpr const char* const name = "kernelcodebook";
        };


    }

//! Kernel Codebook Encoding class.
/*!
  This class implements Kernel Codebook Encoding of features (soft assignments) using K-Means clustering.
  See http://www.robots.ox.ac.uk/~vgg/research/encoding_eval/
*/

class KCEncoding : public Encoding<_internal::KCEncodingTraits>
{
public:
    //! Constructor.
    /*!
      \param k number of bins (and thus features) in encoding
    */
    KCEncoding(int k=256) : _k(k), _maxPooling(false), _tree(nullptr) {}

    //! Clone.
    /*!
      Clones this histogram encoding, returning a pointer to the new one.
    */
    /*Encoding<_internal::HistEncodingTraits>* clone()
    {
        HistEncoding* enc = new HistEncoding();
        enc->_k = _k;
        enc->_usePCA = _usePCA;
        enc->_pcaDim = _pcaDim;
        enc->_pcaBasis = _pcaBasis;

        enc->_centers = _centers;
        enc->_tree = new tet::classification::util::ANNTree<Scalar>(*_tree);

        return enc;
    }*/

    int getK() const { return _k; }
    
    //! Sets number of bins
    /*!
      Sets the number of bins for the encoding.
      \param k number of bins (and thus features) in encoding
    */
    void setK(int k) { _k = k; _computed = false; }

    //! Initialize encoding with descriptor
    /*!
      Computes cluster centers for histogram encoding using passed descriptor
    */
    template <class _DescriptorType>
    void init(const _DescriptorType & descriptor, tet::classification::processing::GetSetDescriptorsBase<_DescriptorType> & descGetter);

    //! Generate histogram encoding from samples
    /*!
      Generates histogram encoding from the samples given. If clusters
      were computed before, the encoding is generated based on these clusters.
      \param samples input samples, one row per sample
      \param recompute if true, clusters are recomputed even if clusters were computed before
      \return histogram of cluster assignments
    */
    FeatureType computeFeatures(SampleType samples) const;

    //! Returns computed cluster centers
    SampleType getCenters() const { return _centers; }
    //! Sets centers from which to find neighbours
    void setCenters(SampleType centers);

    //! Returns FLANN tree
    //tet::classification::util::FLANNTree<Scalar>* getTree() { return _tree; }
    tet::classification::util::ANNTree<Scalar>* getTree() { return _tree; }
    
    // serialization
    tinyxml2::XMLElement* serializeSpec(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc);
    void deserializeSpec(const tinyxml2::XMLElement*);

    friend std::size_t hash_value(const KCEncoding &);

protected:

private:
    int _k;

    Set _samples;

    bool _maxPooling;

    SampleType _centers; // computed cluster centers
    //tet::classification::util::FLANNTree<Scalar> *_tree; // tree for quick neighbour lookup
    tet::classification::util::ANNTree<Scalar> *_tree; // tree for quick neighbour lookup

}; // class KCEncoding

inline std::size_t hash_value(const KCEncoding & enc)
{
    size_t seed = 0;

    boost::hash_combine(seed, enc._k);
    boost::hash_combine(seed, enc._samples);
    boost::hash_combine(seed, enc._maxPooling);
    boost::hash_combine(seed, enc._centers);

    return seed;
}

} // namespace encodings

} // namespace classification

} // namespace tet

#include "impl/kcencoding.hpp"

#endif /* KCENCODING_H */
