#ifndef TOSVMLIGHT_H
#define TOSVMLIGHT_H 

#include <string>
#include <iostream>
#include <fstream>
#include <vector>

#include "boost/lexical_cast.hpp"

#include "eigen3/Eigen/Core"

#include "tet/util/printutils.h"

#include "tet/containers/set.h"

namespace tet
{

namespace classification
{

namespace io
{

template <class _MatType>
std::string toSVMLight(_MatType *data,
                       Eigen::MatrixXi assignments)
{
    std::stringstream result;

    for (size_t i = 0; i < data->rows(); ++i)
    {
        result << assignments(i, 0) << " ";
        for (size_t j = 0; j < data->cols(); ++j)
        {
            if ((*data)(i, j) != 0)
            {
                result << j + 1 << ":" << (*data)(i,j) << " ";
            }
        }
        result << std::endl;
    }

    return result.str();
}

template <class _MatType, class _Scalar>
std::string toSVMLight(const std::vector<_MatType*> & data,
                       Eigen::Matrix<_Scalar, -1, -1> assignments)
{
    std::stringstream result;

    for (size_t i = 0; i < data.size(); ++i)
    {
        result << assignments(i, 0) << " ";
        for (size_t j = 0; j < data.at(i)->rows(); ++j)
        {
            if ((*data.at(i))(j) != 0)
            {
                result << j + 1 << ":" << (*data.at(i))(j) << " ";
            }
        }
        result << std::endl;
    }

    return result.str();
}

template <class _MatType, class _Scalar>
std::string toSVMLight(const std::vector<_MatType*> & data,
                       Eigen::Matrix<_Scalar, -1, 1> assignments)
{
    std::stringstream result;

    for (size_t i = 0; i < data.size(); ++i)
    {
        result << assignments(i, 0) << " ";
        for (size_t j = 0; j < data.at(i)->rows(); ++j)
        {
            if ((*data.at(i))(j) != 0)
            {
                result << j + 1 << ":" << (*data.at(i))(j) << " ";
            }
        }
        result << std::endl;
    }

    return result.str();
}

inline void updateSetWithScores
(tet::classification::containers::Set & set, std::string scoresFilename)
{
    std::ifstream scoresFile;
    scoresFile.open(scoresFilename);

    std::string scoresLine;

    std::vector<tet::classification::containers::Class>::iterator    classIt;
    std::vector<tet::classification::containers::Instance>::iterator instanceIt;
    std::vector<tet::classification::containers::View>::iterator     viewIt;

    for (classIt = set.getClasses().begin(); classIt != set.getClasses().end(); ++classIt)
    {
        for (instanceIt = classIt->getInstances().begin(); instanceIt != classIt->getInstances().end(); ++instanceIt)
        {
            for (viewIt = instanceIt->getViews().begin(); viewIt != instanceIt->getViews().end(); ++viewIt)
            {
                std::getline(scoresFile, scoresLine);
                viewIt->setScore(boost::lexical_cast<float>(scoresLine));
            }
        }
    }
}

template <class _DescriptorType, class _EncoderType>
void updateEncoderWithWeights
(_EncoderType* encoder, std::string weightsFilename)
{
    std::ifstream weightsFile;
    weightsFile.open(weightsFilename);

    std::string weightLine;

    std::vector<typename _EncoderType::Scalar> weights;
    std::getline(weightsFile, weightLine); // bias
    while (std::getline(weightsFile, weightLine))
    {
        weights.push_back(boost::lexical_cast<typename _EncoderType::Scalar>(weightLine));
    }
    encoder->setFeatureWeights(weights);
}

inline void combineResultsWithScores(std::string testFilename, std::string scoresFilename, std::string outFilename)
{
    std::cout << "[combineResultsWithScore] Combining " << testFilename << " and " 
                                                        << scoresFilename << " into " 
                                                        << outFilename << "." << std::endl;

    std::vector<std::pair<double, std::string> > filesWithScores;

    std::ifstream testFile;
    testFile.open(testFilename);

    std::ifstream scoresFile;
    scoresFile.open(scoresFilename);

    for (std::string testLine, scoresLine; std::getline(testFile, testLine) && std::getline(scoresFile, scoresLine); )
    {
        filesWithScores.push_back(std::pair<double, std::string>(boost::lexical_cast<double>(scoresLine), testLine));
    }
    
    std::sort(filesWithScores.begin(), filesWithScores.end());

    std::ofstream outFile;
    outFile.open(outFilename);
    outFile << "file,score" << std::endl;

    for (auto it = filesWithScores.begin(); it != filesWithScores.end(); ++it)
    {
        outFile << it->second << "," << it->first << std::endl;
    }

    outFile.close();
}

} // namespace io

} // namespace classification

} // namespace tet

#endif /* TOSVMLIGHT_H */
