#ifndef DATAPROCESSING_H
#define DATAPROCESSING_H 

#include <boost/unordered_map.hpp>

#include "tet/util/fileutils.h"

#include "tet/processing/serialization.h"

#include "tet/containers/view.h"
#include "tet/containers/instance.h"
#include "tet/containers/class.h"
#include "tet/containers/set.h"

using namespace tet::classification::containers;

namespace tet
{

namespace classification
{

namespace processing
{

// View functions
template <class _DescriptorType>
struct GetViewDescriptorBase
{
    virtual typename _DescriptorType::FeatureType operator()
    (const View &, const _DescriptorType &) = 0;
};

template <class _DescriptorType>
struct GetViewDescriptor : public GetViewDescriptorBase<_DescriptorType>
{
    typename _DescriptorType::FeatureType operator()
    (const View & view, const _DescriptorType & desc)
    {
        typename _DescriptorType::FeatureType feature = desc.getDescriptors(*(view.getInput<typename _DescriptorType::MatType>()));
        view.unloadInput();
        return feature;
    }
};
    
template <class _DescriptorType>
struct GetViewDescriptorCached : public GetViewDescriptorBase<_DescriptorType>
{
    GetViewDescriptorCached(const std::string & cacheDir) : _cacheDir(cacheDir) {}

    typename _DescriptorType::FeatureType operator()
    (const View & view, const _DescriptorType & desc)
    {
        size_t seed = 0;
        boost::hash_combine(seed, view);
        boost::hash_combine(seed, desc);

        // DEBUG
        //std::cout << view.getFilename() << std::endl;

        // check if stored on disk
        std::stringstream fn;
        fn << _cacheDir << "/viewdesc" << seed << ".bin";
        typename _DescriptorType::FeatureType feature;
        if (util::file_exists(fn.str()))
        {
            feature = deserialize<typename _DescriptorType::FeatureType>(fn.str());
        } else {
            feature = desc.getDescriptors(*(view.getInput<typename _DescriptorType::MatType>()));
            view.unloadInput();
            serialize<typename _DescriptorType::FeatureType>(feature, fn.str());
        }
        return feature;
    }

private:
    std::string _cacheDir;
};

template <class _DescriptorType, class _EncoderType>
struct GetViewEncodingBase
{
    virtual typename _EncoderType::FeatureType* operator()
    (const View &, const _DescriptorType &, const _EncoderType &) = 0;
};

template <class _DescriptorType, class _EncoderType>
struct GetViewEncoding : public GetViewEncodingBase<_DescriptorType, _EncoderType>
{
    typename _EncoderType::FeatureType* operator()
    (const View & view, const _DescriptorType & desc, const _EncoderType & enc)
    {
        typename _EncoderType::FeatureType* feature = 
            enc.computeFeatures(desc.getDescriptors(*(view.getInput<typename _DescriptorType::MatType>())));
        view.unloadInput();

        return feature;
    }
};

template <class _DescriptorType, class _EncoderType>
struct GetViewEncodingCached : public GetViewEncodingBase<_DescriptorType, _EncoderType>
{
    GetViewEncodingCached(std::string cacheDir) : _cacheDir(cacheDir), _getViewDescriptor(cacheDir) {}

    typename _EncoderType::FeatureType* operator()
    (const View & view, const _DescriptorType & desc, const _EncoderType & enc)
    {
        size_t seed = 0;
        boost::hash_combine(seed, view);
        boost::hash_combine(seed, desc);
        boost::hash_combine(seed, enc);

        std::cout << view.getFilename() << std::endl;

        // check if stored on disk
        std::stringstream fn;
        fn << _cacheDir << "/viewenc" << seed << ".bin";
        typename _EncoderType::FeatureType* feature = new typename _EncoderType::FeatureType;
        if (util::file_exists(fn.str()))
        {
            *feature = deserialize<typename _EncoderType::FeatureType>(fn.str());
        } else {
            *feature = enc.computeFeatures(_getViewDescriptor(view, desc));
            serialize<typename _EncoderType::FeatureType>(*feature, fn.str());
        }

        return feature;
    }

private:
    std::string _cacheDir;
    GetViewDescriptorCached<_DescriptorType> _getViewDescriptor;
};

// Instance functions
template <class _DescriptorType>
struct GetInstanceDescriptorsBase
{
    virtual std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Instance & inst, const _DescriptorType & desc) = 0;
};

template <class _DescriptorType>
struct GetInstanceDescriptors : public GetInstanceDescriptorsBase<_DescriptorType>
{
    GetInstanceDescriptors() : _getViewDescriptor() {}

    std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Instance & inst, const _DescriptorType & desc)
    {
        std::vector<typename _DescriptorType::FeatureType*> descriptors(inst.size());

        for (size_t viewIdx = 0;
             viewIdx < inst.getViews().size();
             ++viewIdx)
        {
            descriptors[viewIdx] = new typename _DescriptorType::FeatureType;
            *(descriptors[viewIdx]) = _getViewDescriptor(inst.getViews()[viewIdx], desc);
        }

        return descriptors;
    }

private:
    GetViewDescriptor<_DescriptorType> _getViewDescriptor;
};

template <class _DescriptorType>
struct GetInstanceDescriptorsCached : public GetInstanceDescriptorsBase<_DescriptorType>
{
    GetInstanceDescriptorsCached(std::string cacheDir) : _getViewDescriptor(cacheDir) {}

    std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Instance & inst, const _DescriptorType & desc)
    {
        std::vector<typename _DescriptorType::FeatureType*> descriptors(inst.size());

        for (size_t viewIdx = 0;
             viewIdx < inst.getViews().size();
             ++viewIdx)
        {
            descriptors[viewIdx] = new typename _DescriptorType::FeatureType;
            *(descriptors[viewIdx]) = _getViewDescriptor(inst.getViews()[viewIdx], desc);
        }

        return descriptors;
    }

private:
    GetViewDescriptorCached<_DescriptorType> _getViewDescriptor;
};

template <class _DescriptorType, class _EncoderType>
struct GetInstanceEncodingsBase
{
    virtual std::vector<typename _EncoderType::FeatureType*> operator()
    (const Instance & inst, const _DescriptorType & desc, const _EncoderType & enc) = 0;
};

template <class _DescriptorType, class _EncoderType>
struct GetInstanceEncodings : public GetInstanceEncodingsBase<_DescriptorType, _EncoderType>
{
    GetInstanceEncodings() : _getViewEncoding() {}

    std::vector<typename _EncoderType::FeatureType*> operator()
    (const Instance & inst, const _DescriptorType & desc, const _EncoderType & enc)
    {
        std::vector<typename _EncoderType::FeatureType*> encodings(inst.size());

        for (size_t viewIdx = 0;
             viewIdx < inst.getViews().size();
             ++viewIdx)
        {
            encodings[viewIdx] = _getViewEncoding(inst.getViews()[viewIdx], desc, enc);
        }

        return encodings;
    }

private:
    GetViewEncoding<_DescriptorType, _EncoderType> _getViewEncoding;
};

template <class _DescriptorType, class _EncoderType>
struct GetInstanceEncodingsCached : public GetInstanceEncodingsBase<_DescriptorType, _EncoderType>
{
    GetInstanceEncodingsCached(std::string cacheDir) : _getViewEncoding(cacheDir) {}

    std::vector<typename _EncoderType::FeatureType*> operator()
    (const Instance & inst, const _DescriptorType & desc, const _EncoderType & enc)
    {
        std::vector<typename _EncoderType::FeatureType*> encodings(inst.size());

        for (size_t viewIdx = 0;
             viewIdx < inst.getViews().size();
             ++viewIdx)
        {
            encodings[viewIdx] = _getViewEncoding(inst.getViews()[viewIdx], desc, enc);
        }

        return encodings;
    }

private:
    GetViewEncodingCached<_DescriptorType, _EncoderType> _getViewEncoding;

};

// Class functions
template <class _DescriptorType>
struct GetClassDescriptorsBase
{
    virtual std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Class &, const _DescriptorType &) = 0;
};

template <class _DescriptorType>
struct GetClassDescriptors : public GetClassDescriptorsBase<_DescriptorType>
{
    GetClassDescriptors() : _getInstanceDescriptors() {}

    std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Class & c, const _DescriptorType & desc)
    {
        std::vector<typename _DescriptorType::FeatureType*> descriptors(c.size());

        size_t curIdx = 0;
        for (size_t instanceIdx = 0;
             instanceIdx < c.getInstances().size();
             ++instanceIdx)
        {
            const std::vector<typename _DescriptorType::FeatureType*> instanceDescriptors = 
                _getInstanceDescriptors(c.getInstances()[instanceIdx], desc);

            for (size_t i = curIdx; i < curIdx + c.getInstances()[instanceIdx].size(); ++i)
            {
                descriptors[i] = instanceDescriptors[i - curIdx];
            }
            curIdx += c.getInstances()[instanceIdx].size();

        }

        return descriptors;
    }

private:
    GetInstanceDescriptors<_DescriptorType> _getInstanceDescriptors;

};

template <class _DescriptorType>
struct GetClassDescriptorsCached : public GetClassDescriptorsBase<_DescriptorType>
{
    GetClassDescriptorsCached(std::string cacheDir) : _getInstanceDescriptors(cacheDir) {}

    std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Class & c, const _DescriptorType & desc)
    {
        std::vector<typename _DescriptorType::FeatureType*> descriptors(c.size());

        size_t curIdx = 0;
        for (size_t instanceIdx = 0;
             instanceIdx < c.getInstances().size();
             ++instanceIdx)
        {
            const std::vector<typename _DescriptorType::FeatureType*> instanceDescriptors = 
                _getInstanceDescriptors(c.getInstances()[instanceIdx], desc);

            for (size_t i = curIdx; i < curIdx + c.getInstances()[instanceIdx].size(); ++i)
            {
                descriptors[i] = instanceDescriptors[i - curIdx];
            }
            curIdx += c.getInstances()[instanceIdx].size();

        }

        return descriptors;
    }

private:
    GetInstanceDescriptorsCached<_DescriptorType> _getInstanceDescriptors;

};

template <class _DescriptorType, class _EncoderType>
struct GetClassEncodingsBase
{
    virtual std::vector<typename _EncoderType::FeatureType*> operator()
    (const Class &, const _DescriptorType &, const _EncoderType &) = 0;
};

template <class _DescriptorType, class _EncoderType>
struct GetClassEncodings : public GetClassEncodingsBase<_DescriptorType, _EncoderType>
{
    GetClassEncodings() : _getInstanceEncodings() {}

    std::vector<typename _EncoderType::FeatureType*> operator()
    (const Class & c, const _DescriptorType & desc, const _EncoderType & enc)
    {
        std::vector<typename _EncoderType::FeatureType*> encodings;

        for (size_t instanceIdx = 0;
             instanceIdx < c.getInstances().size();
             ++instanceIdx)
        {
            const std::vector<typename _EncoderType::FeatureType*> instanceEncodings = 
                _getInstanceEncodings(c.getInstances()[instanceIdx], desc, enc);

            encodings.insert(encodings.end(), instanceEncodings.begin(), instanceEncodings.end());
        }

        return encodings;
    }

private:
    GetInstanceEncodings<_DescriptorType, _EncoderType> _getInstanceEncodings;

};

template <class _DescriptorType, class _EncoderType>
struct GetClassEncodingsCached : public GetClassEncodingsBase<_DescriptorType, _EncoderType>
{
    GetClassEncodingsCached(std::string cacheDir) : _getInstanceEncodings(cacheDir) {}

    std::vector<typename _EncoderType::FeatureType*> operator()
    (const Class & c, const _DescriptorType & desc, const _EncoderType & enc)
    {
        std::vector<typename _EncoderType::FeatureType*> encodings;

        for (size_t instanceIdx = 0;
             instanceIdx < c.getInstances().size();
             ++instanceIdx)
        {
            const std::vector<typename _EncoderType::FeatureType*> instanceEncodings = 
                _getInstanceEncodings(c.getInstances()[instanceIdx], desc, enc);

            encodings.insert(encodings.end(), instanceEncodings.begin(), instanceEncodings.end());
        }

        return encodings;
    }

private:
    GetInstanceEncodingsCached<_DescriptorType, _EncoderType> _getInstanceEncodings;

};

// Set functions
template <class _DescriptorType>
struct GetSetDescriptorsBase
{
    virtual std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Set &, const _DescriptorType &) = 0;
};

template <class _DescriptorType>
struct GetSetDescriptors: public GetSetDescriptorsBase<_DescriptorType>
{
    GetSetDescriptors() : _getClassDescriptors() {}

    std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Set & set, const _DescriptorType & desc)
    {
        std::vector<typename _DescriptorType::FeatureType*> descriptors;

        for (size_t classIdx = 0;
             classIdx < set.getClasses().size();
             ++classIdx)
        {
            const std::vector<typename _DescriptorType::FeatureType*> classDescriptors =
                _getClassDescriptors(set.getClasses()[classIdx], desc);
            
            descriptors.insert(descriptors.end(), classDescriptors.begin(), classDescriptors.end());
        }

        return descriptors;
    }

private:
    GetClassDescriptors<_DescriptorType> _getClassDescriptors;

};

template <class _DescriptorType>
struct GetSetDescriptorsCached : public GetSetDescriptorsBase<_DescriptorType>
{
    GetSetDescriptorsCached(std::string cacheDir) : _getClassDescriptors(cacheDir) {}

    std::vector<typename _DescriptorType::FeatureType*> operator()
    (const Set & set, const _DescriptorType & desc)
    {
        std::vector<typename _DescriptorType::FeatureType*> descriptors;

        for (size_t classIdx = 0;
             classIdx < set.getClasses().size();
             ++classIdx)
        {
            const std::vector<typename _DescriptorType::FeatureType*> classDescriptors =
                _getClassDescriptors(set.getClasses()[classIdx], desc);
            
            descriptors.insert(descriptors.end(), classDescriptors.begin(), classDescriptors.end());
        }

        return descriptors;
    }

private:
    GetClassDescriptorsCached<_DescriptorType> _getClassDescriptors;

};

template <class _DescriptorType, class _EncoderType>
struct GetSetEncodingsBase
{
    virtual std::vector<typename _EncoderType::FeatureType*> operator()
    (const Set &, const _DescriptorType &, const _EncoderType &) = 0;
};

template <class _DescriptorType, class _EncoderType>
struct GetSetEncodings : public GetSetEncodingsBase<_DescriptorType, _EncoderType>
{
    GetSetEncodings(std::string cacheDir) : _getClassEncodings(cacheDir) {}

    std::vector<typename _EncoderType::FeatureType*> operator()
    (const Set & set, const _DescriptorType & desc, const _EncoderType & enc)
    {
        std::vector<typename _EncoderType::FeatureType*> encodings;

        for (size_t classIdx = 0;
             classIdx < set.getClasses().size();
             ++classIdx)
        {
            const std::vector<typename _EncoderType::FeatureType*> classEncodings =
                _getClassEncodings(set.getClasses()[classIdx], desc, enc);
            
            encodings.insert(encodings.end(), classEncodings.begin(), classEncodings.end());
        }

        return encodings;
    }

private:
    GetClassEncodings<_DescriptorType, _EncoderType> _getClassEncodings;

};

template <class _DescriptorType, class _EncoderType>
struct GetSetEncodingsCached : public GetSetEncodingsBase<_DescriptorType, _EncoderType>
{
    GetSetEncodingsCached(std::string cacheDir) : _getClassEncodings(cacheDir) {}

    std::vector<typename _EncoderType::FeatureType*> operator()
    (const Set & set, const _DescriptorType & desc, const _EncoderType & enc)
    {
        std::vector<typename _EncoderType::FeatureType*> encodings;

        for (size_t classIdx = 0;
             classIdx < set.getClasses().size();
             ++classIdx)
        {
            const std::vector<typename _EncoderType::FeatureType*> classEncodings =
                _getClassEncodings(set.getClasses()[classIdx], desc, enc);
            
            encodings.insert(encodings.end(), classEncodings.begin(), classEncodings.end());
        }

        return encodings;
    }

private:
    GetClassEncodingsCached<_DescriptorType, _EncoderType> _getClassEncodings;

};

} // namespace processing

} // namespace classification

} // namespace tet

#endif /* DATAPROCESSING_H */
