#!/usr/bin/env python
import sys
import os
import subprocess
import argparse
import copy

import tet.storage
import tet.data
import tet.run
from tet.descriptors.descriptorfactory import DescriptorFactory
from tet.encodings.encodingfactory     import EncodingFactory
from tet.kernels.kernelfactory         import KernelFactory

RUN_CONFIG_DIR = '/path/to/xml/run/file/storage/dir'
RUN_OUTPUT_DIR = '/path/to/xml/result/file/storage/dir'
DEFAULT_CACHE_DIR = '/path/to/cache/storage/dir'

## Creates a ClassifierRun instance for a specific Bing class, and returns it
#  @param className The class name
def createClassifierRun(className, cacheDir, descriptor, encoding, kernel):
    # get Bing class and instances for positive
    myClass = tet.storage.MyClass(className, tet.data.getBingClassID(className))
    myClass.instances = tet.data.getBingInstances(myClass.idx)

    myNegClasses = [cl for cl in tet.data.getBingClasses() if cl.name != className]
    for cl in myNegClasses:
        cl.instances = tet.data.getBingInstances(cl.idx)

    myClassTest = tet.storage.MyClass(className, tet.data.getTDWClassID(className))
    myClassTest.instances = tet.data.getTDWInstances(tet.data.getTDWClassID(className))
    for instance in myClassTest.instances:
        instance.views = tet.data.getTDWViews(instance.idx)

    run = tet.run.ClassifierRun()
    run.name       = "{0}_classification".format(className)
    run.cacheDir   = cacheDir
    run.kernel     = kernel

    run.setDescriptor(descriptor)
    run.setEncoding(encoding)
    encoding.init(run)

    run.addPositives([myClass])
    run.addNegatives(myNegClasses)
    run.addTests([myClassTest])

    return run

def runClassification(settings, descriptor, encoding, kernel):
    # create run input xml
    run = createClassifierRun(settings.classname, settings.cachedir, descriptor, encoding, kernel)
    runInputFilename = os.path.join(RUN_CONFIG_DIR, settings.classname, "{0}_{1}_{2}_{3}.xml".format(settings.descriptor, settings.encoding, settings.kernel, run.name))
    if not os.path.exists(os.path.dirname(runInputFilename)):
        os.mkdir(os.path.dirname(runInputFilename))
    run.writeToFile(runInputFilename)

    # create run output xml and run
    runOutputFilename = os.path.join(RUN_OUTPUT_DIR, settings.classname, "{0}_{1}_{2}_{3}.xml".format(settings.descriptor, settings.encoding, settings.kernel, run.name))
    if not os.path.exists(os.path.dirname(runOutputFilename)):
        os.mkdir(os.path.dirname(runOutputFilename))

    tet.run.runClassifierFromXML(runInputFilename, runOutputFilename, settings.encoding, settings.descriptor, ['-c', '1'])

    return runOutputFilename
    

if __name__=="__main__":
    parser = argparse.ArgumentParser(description='Run 3D model classification for given class and settings.')
    parser.add_argument('classname', help='name of the class')
    parser.add_argument('--descriptor', choices=['hog', 'cnn'], required=True)
    parser.add_argument('--encoding', choices=['kc', 'iden'], required=True)
    parser.add_argument('--cachedir', help='cache storage directory', default=DEFAULT_CACHE_DIR)
    parser.add_argument('--kernel', help='type of kernel', choices=['linear', 'polynomial', 'rbf', 'sigmoid'], default='linear')
    parser.add_argument('--r', help='constant value for polynomial and sigmoid kernel', default=1)
    parser.add_argument('--a', help='scaling value for polynomial and sigmoid kernel', default=1)
    parser.add_argument('--d', help='dimension of polynomial kernel', default=2)
    parser.add_argument('--gamma', help='gamma value for RBF kernel', default=0.001)
    settings = parser.parse_args()

    descriptor = DescriptorFactory().get(settings.descriptor)
    encoding = EncodingFactory().get(settings.encoding)
    kernel = KernelFactory().get(settings.kernel)
    kernel.fromSettings(settings)

    runClassification(settings, descriptor, encoding, kernel)

    print "Done with classification for class {0}.".format(settings.classname)
