#!/usr/bin/env python
import sys
import os
import subprocess
import argparse

import copy
import numpy

import tet.storage
import tet.data
import tet.run
from tet.descriptors.descriptorfactory import DescriptorFactory
from tet.encodings.encodingfactory     import EncodingFactory
from tet.kernels.kernelfactory         import KernelFactory

from tet.encodings.kcencoding          import KCEncoding

from tet.util.xmlUtils                 import getInstancesAboveThreshold

RUN_CONFIG_DIR = '/path/to/xml/run/file/storage/dir'
RUN_OUTPUT_DIR = '/path/to/xml/result/file/storage/dir'
DEFAULT_CACHE_DIR = '/path/to/cache/storage/dir'

## Creates a Run instance for a specific 3DW class and view index, and returns it.
#  @param className The class name
#  @param viewID The view index
#  @param genCustom (optional) True if custom set should be generated from ALL
#         views. Defaults to False.
def createViewRun(className, cacheDir, descriptor, encoding, kernel, viewID, filteredInstances, genCustom=False):
    myClass = tet.storage.MyClass(className, tet.data.getTDWClassID(className))
    myInstances = tet.data.getTDWInstances(myClass.idx)

    if (filteredInstances is not None):
        myInstances = [instance for instance in myInstances if instance.file_idx in filteredInstances]

    N = len(tet.data.getTDWViews(myInstances[0].idx))

    negLID = (viewID - (N / 4)) % N
    negRID = (viewID + (N / 4)) % N
    #negMID = (viewID + (N / 2)) % N

    myClassTest = copy.deepcopy(myClass)
    myClassTest.instances = tet.data.getBingInstances(tet.data.getBingClassID(className))

    myClassNeg  = copy.deepcopy(myClass)
    myClassPos  = copy.deepcopy(myClass)
    if genCustom:
        myClassCustom = copy.deepcopy(myClass)

    run = tet.run.ClassifierRun()
    run.name     = "{0}_{1:02d}".format(className, viewID)
    run.cacheDir = cacheDir
    run.kernel   = kernel


    for instance in myInstances:
        posInstance = copy.deepcopy(instance)
        negInstance = copy.deepcopy(instance)
        if genCustom:
            customInstance = copy.deepcopy(instance)

        views = tet.data.getTDWViews(instance.idx)
        offsetViewID = (viewID - instance.align_offset) % N
        offsetNegLID = (negLID - instance.align_offset) % N
        offsetNegRID = (negRID - instance.align_offset) % N
        #offsetNegMID = (negMID - instance.align_offset) % N

        posViews  = list(numpy.take(views, range(offsetViewID - 1, offsetViewID + 2), mode='wrap'))
        negViews  = list(numpy.take(views, range(offsetNegLID - 1, offsetNegLID + 2), mode='wrap'))
        negViews += list(numpy.take(views, range(offsetNegRID - 1, offsetNegRID + 2), mode='wrap'))
        #negViews += list(numpy.take(views, range(offsetNegMID - 1, offsetNegMID + 2), mode='wrap'))
        if genCustom:
            customViews = list(views)

        posInstance.views = posViews
        negInstance.views = negViews
        if genCustom:
            customInstance.views = customViews

        myClassPos.instances.append(posInstance)
        myClassNeg.instances.append(negInstance)
        if genCustom:
            myClassCustom.instances.append(customInstance)

    run.addPositives([myClassPos])
    run.addNegatives([myClassNeg])
    run.addTests([myClassTest])

    run.setDescriptor(descriptor)
    run.setEncoding(encoding)

    if genCustom:
        run.addCustoms([myClassCustom])

    return run

def runViewClassification(settings, descriptor, encoding, kernel):
    # use filter file if specified
    if (settings.filter is not None):
        filteredInstances = getInstancesAboveThreshold(settings.filter, settings.threshold)
    else:
        filteredInstances = None

    # special case for KC codebook kernel (all views need to have same
    # set from which K centers are created)
    if isinstance(encoding, KCEncoding):
        myClass = tet.storage.MyClass(settings.classname, tet.data.getTDWClassID(settings.classname))
        myInstances = tet.data.getTDWInstances(myClass.idx)

        if (filteredInstances is not None):
            myInstances = [instance for instance in myInstances if instance.file_idx in filteredInstances]

        for instance in myInstances:
            views = tet.data.getTDWViews(instance.idx)
            instance.views = views
            myClass.instances.append(instance)
        encoding.samples = [myClass]

    # create all runs and execute them
    for viewID in range(0, 36):
        run = createViewRun(settings.classname, settings.cachedir, descriptor, encoding, kernel, viewID, filteredInstances)
        runInputFilename = os.path.join(RUN_CONFIG_DIR, settings.classname, "{0}_{1}_{2}_{3}.xml".format(settings.descriptor, settings.encoding, settings.kernel, run.name))
        if not os.path.exists(os.path.dirname(runInputFilename)):
            os.mkdir(os.path.dirname(runInputFilename))
        run.writeToFile(runInputFilename)

        runOutputFilename = os.path.join(RUN_OUTPUT_DIR, settings.classname, "{0}_{1}_{2}_{3}.xml".format(settings.descriptor, settings.encoding, settings.kernel, run.name))
        if not os.path.exists(os.path.dirname(runOutputFilename)):
            os.mkdir(os.path.dirname(runOutputFilename))

        tet.run.runClassifierFromXML(runInputFilename, runOutputFilename, settings.encoding, settings.descriptor, ['-c', '1'])

if __name__=="__main__":
    parser = argparse.ArgumentParser(description='Run 3D model classification for given class and settings.')
    parser.add_argument('classname', help='name of the class')
    parser.add_argument('--descriptor', choices=['hog', 'cnn'], required=True)
    parser.add_argument('--encoding', choices=['kc', 'iden'], required=True)
    parser.add_argument('--cachedir', help='cache storage directory', default=DEFAULT_CACHE_DIR)
    parser.add_argument('--kernel', help='type of kernel', choices=['linear', 'polynomial', 'rbf', 'sigmoid'], default='linear')
    parser.add_argument('--r', help='constant value for polynomial and sigmoid kernel', default=1)
    parser.add_argument('--a', help='scaling value for polynomial and sigmoid kernel', default=1)
    parser.add_argument('--d', help='dimension of polynomial kernel', default=2)
    parser.add_argument('--gamma', help='gamma value for RBF kernel', default=0.001)
    parser.add_argument('--filter', help='only use instances scoring higher than threshold from XML file', default=None)
    parser.add_argument('--threshold', help='threshold for use with filter', default=0.0)
    settings = parser.parse_args()

    descriptor = DescriptorFactory().get(settings.descriptor)
    encoding = EncodingFactory().get(settings.encoding)
    kernel = KernelFactory().get(settings.kernel)
    kernel.fromSettings(settings)

    runViewClassification(settings, descriptor, encoding, kernel)

    print "Done with all view classifiers for class {0}.".format(settings.classname)
