from lxml import etree
import StringIO

import os
import subprocess

CLASSIFIER = '/path/to/classifier/executable'
REGRESSOR = '/path/to/regressor/executable'

## Models a Classifier Run
class ClassifierRun:
    def __init__(self):
        self.positives = []
        self.negatives = []
        self.tests = []
        self.customs = []
        self.kernel = None
        self.cacheDir = ""
        self.descriptor = None
        self.encoding = None
        self.name = "default"
        self.c = None
        self.evaluations = []

    ## Adds list of positive samples to the positives
    #  @param positives The positives to add to the training set
    def addPositives(self, positives):
        self.positives += positives

    ## Adds list of negative samples to the negatives
    #  @param negatives The negatives to add to the training set
    def addNegatives(self, negatives):
        self.negatives += negatives

    ## Adds list of test samples to the tests
    #  @param tests The tests to add to the test set
    def addTests(self, tests):
        self.tests += tests

    ## Adds list of custom samples to the customs
    #  @param customs The customs to add to the custom set
    def addCustoms(self, customs):
        self.customs += customs

    ## Sets descriptor for this run
    #  @param descriptor Descriptor to set
    def setDescriptor(self, descriptor):
        self.descriptor = descriptor

    ## Sets encoding for this run
    #  @param encoding Encoding to set
    def setEncoding(self, encoding):
        self.encoding = encoding

    ## Adds evaluation method to use
    #  @param evalStr String representation of evaluation
    def addEvaluation(self, evalStr):
        self.evaluations.append(evalStr)


    def getClusteringSamples(self):
        return self.negatives

    ## Converts this run to an XML element and returns it
    def toXML(self):
        root = etree.Element("file")
        settings = etree.SubElement(root, "settings")
        name = etree.SubElement(settings, "name")
        name.text = self.name
        if self.c:
            c = etree.SubElement(settings, "c")
            c.text = str(self.c)
        positives = etree.SubElement(settings, "positives")
        posSet = etree.SubElement(positives, "set")
        for pos in self.positives:
            posSet.append(pos.toXML())
        negatives = etree.SubElement(settings, "negatives")
        negSet = etree.SubElement(negatives, "set")
        for neg in self.negatives:
            negSet.append(neg.toXML())
        tests = etree.SubElement(settings, "tests")
        testSet = etree.SubElement(tests, "set")
        for test in self.tests:
            testSet.append(test.toXML())
        customs = etree.SubElement(settings, "custom")
        cusSet = etree.SubElement(customs, "set")
        for cus in self.customs:
            cusSet.append(cus.toXML())

        #defaultSettings = etree.parse("settings.xml").getroot()
        #for setting in defaultSettings:
        #    settings.append(setting)

        cacheDir = etree.SubElement(settings, "cacheDir")
        cacheDir.text = self.cacheDir

        settings.append(self.kernel.toXML())

        for evaluation in self.evaluations:
            evalTag = etree.SubElement(settings, "eval")
            evalTag.attrib["type"] = evaluation

        settings.append(self.descriptor.toXML())
        settings.append(self.encoding.toXML())

        return root

    ## Writes this run to an XML file
    #  @param filename The filename to write the XML output to
    def writeToFile(self, filename):
        f = open(filename, 'w')
        f.write(self.toString())
        f.close()

    ## Converts this run to an XML string
    def toString(self):
        parser = etree.XMLParser(remove_blank_text=True)
        file_obj = StringIO.StringIO(etree.tostring(self.toXML()))
        reparsed = etree.parse(file_obj, parser)
        return etree.tostring(reparsed, pretty_print=True)

## Models a Regressor Run
class RegressorRun:
    def __init__(self):
        self.samples = []
        self.tests = []
        self.name = "default"
        self.kernel = None
        self.cacheDir = ""
        self.descriptor = None
        self.encoding = None
        self.c = None
        self.eps = None
        self.evaluations = []

    ## Adds list of samples to the training set
    #  @param samples The samples to be added to the training set (list)
    def addSamples(self, samples):
        self.samples += samples

    ## Adds list of test cases to the test set
    #  @param tests The test cases to be added to the test set (list)
    def addTests(self, tests):
        self.tests += tests

    ## Sets descriptor for this run
    #  @param descriptor Descriptor to set
    def setDescriptor(self, descriptor):
        self.descriptor = descriptor

    ## Sets encoding for this run
    #  @param encoding Encoding to set
    def setEncoding(self, encoding):
        self.encoding = encoding
        self.encoding.init(self)

    ## Adds evaluation method to use
    #  @param evalStr String representation of evaluation
    def addEvaluation(self, evalStr):
        self.evaluations.append(evalStr)

    def getClusteringSamples(self):
        return self.samples

    ## Converts this run to an XML element and returns it
    def toXML(self):
        root = etree.Element("file")
        settings = etree.SubElement(root, "settings")
        name = etree.SubElement(settings, "name")
        name.text = self.name
        if self.c:
            c = etree.SubElement(settings, "c")
            c.text = str(self.c)
        if self.eps:
            eps = etree.SubElement(settings, "eps")
            eps.text = str(self.eps)
        samples = etree.SubElement(settings, "samples")
        samplesSet = etree.SubElement(samples, "set")
        for sample in self.samples:
            samplesSet.append(sample.toXML())
        tests = etree.SubElement(settings, "tests")
        testsSet = etree.SubElement(tests, "set")
        for test in self.tests:
            testsSet.append(test.toXML())

        #defaultSettings = etree.parse("settings.xml").getroot()
        #for setting in defaultSettings:
        #    settings.append(setting)
        cacheDir = etree.SubElement(settings, "cacheDir")
        cacheDir.text = self.cacheDir

        settings.append(self.kernel.toXML())
        settings.append(self.descriptor.toXML())
        settings.append(self.encoding.toXML())

        for evaluation in self.evaluations:
            evalTag = etree.SubElement(settings, "eval")
            evalTag.attrib["type"] = evaluation

        return root

    ## Writes this run to an XML file
    #  @param filename The filename to write the XML output to
    def writeToFile(self, filename):
        f = open(filename, 'w')
        f.write(self.toString())
        f.close()

    ## Converts this run to an XML string
    def toString(self):
        parser = etree.XMLParser(remove_blank_text=True)
        file_obj = StringIO.StringIO(etree.tostring(self.toXML()))
        reparsed = etree.parse(file_obj, parser)
        return etree.tostring(reparsed, pretty_print=True)

def runClassifierFromXML(xmlFile, outFile, enc, desc, extraArgs=[]):
    args = [CLASSIFIER, '-s', xmlFile, '-o', outFile, '-e', enc, '-d', desc]
    args = args + extraArgs

    print ' '.join(args[1:])

    p = subprocess.Popen(args)
    p.wait()

def runRegressorFromXML(xmlFile, outFile, enc, desc, extraArgs):
    args = [REGRESSOR, '-s', xmlFile, '-o', outFile, '-e', enc, '-d', desc]
    args = args + extraArgs

    print args

    p = subprocess.Popen(args)
    p.wait()

def getEvalMetricFromXML(xmlFile, mType):
    root = etree.parse(xmlFile).getroot()
    metricTag = root.find(".//eval[@type='{0}']".format(mType))
    if metricTag is not None:
        return float(metricTag.text)
