#include <iostream>
#include <fstream>
#include <string>

#include "boost/program_options.hpp"

#include "tet/encodings/identityencoding.h"
#include "tet/encodings/kcencoding.h"
#include "tet/descriptors/cnndescriptor.h"
#include "tet/descriptors/hogdescriptor.h"

#include "tet/app/runsettings.h"
#include "tet/app/classifierapp.h"

#include "tet/app/appfactory.h"

namespace {
    const size_t ERROR_IN_OPTIONS = 1;
    const size_t SUCCESS = 0;
}

void usage(char* execName, const boost::program_options::options_description &options)
{
    std::cout << "Usage: " << execName << " [options]" << std::endl;
    std::cout << options;
}

size_t parseArguments(int argc, char* argv[], tet::classification::app::RunSettings &settings)
{
    
    // setting up program options
    namespace po = boost::program_options;

    po::options_description desc("Options");
    desc.add_options()
        ("settings,s", po::value<std::string>(), "Run settings (XML)")
        ("output,o",   po::value<std::string>(), "Output file (XML)")
        ("enc,e",      po::value<std::string>(), "Encoding [kc|iden]")
        ("desc,d",     po::value<std::string>(), "Descriptor [hog|cnn]")
        ("cache,c",    po::value<bool>(),        "Use cached encodings (NOTE: reset is manual by deleting .enc.bin files!)");

    po::variables_map vm;
    try
    {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);

        if (vm.count("help"))
        {
            usage(argv[0], desc);
            return SUCCESS;
        }

        if (vm.count("settings"))
        {
            settings.setSettingsFile(vm["settings"].as<std::string>());
        } else {
            std::cerr << "[Main | Parsing commands] Error: specify settings file" << std::endl;
            return ERROR_IN_OPTIONS;
        }

        if (vm.count("output"))
        {
            settings.setResultsFile(vm["output"].as<std::string>());
        } else {
            std::cerr << "[Main | Parsing commands] Error: specify output file" << std::endl;
            return ERROR_IN_OPTIONS;
        }

        if (vm.count("enc"))
        {
            settings.setEncoding(vm["enc"].as<std::string>());
        } else {
            std::cerr << "[Main | Parsing commands] Error: specify encoding type" << std::endl;
            return ERROR_IN_OPTIONS;
        }

        if (vm.count("desc"))
        {
            settings.setDescriptor(vm["desc"].as<std::string>());
        } else {
            std::cerr << "[Main | Parsing commands] Error: specify descriptor type" << std::endl;
            return ERROR_IN_OPTIONS;
        }

        if (vm.count("cache"))
        {
            settings.setUseCached(vm["cache"].as<bool>());
        }

    }
    catch (po::error& e)
    {
        std::cerr << "[Main | Parsing commands] Error: " << e.what() << std::endl;
        usage(argv[0], desc);
        return ERROR_IN_OPTIONS;
    }

    return SUCCESS;

}

int main(int argc, char* argv[])
{
    tet::classification::app::RunSettings settings;
    size_t parseResult = parseArguments(argc, argv, settings);
    if (parseResult != SUCCESS)
    {
        return parseResult;
    }

    tet::classification::app::AppFactory appFactory;
    tet::classification::app::AppBase * app = appFactory.createApp("class", settings.getDescriptor(), settings.getEncoding());

    app->readFile(settings.getSettingsFile());
    app->run();
    app->writeFile(settings.getResultsFile());

}
