#include "tet/classifiers/svm.h"
#include "tet/util/kernelutils.h"

namespace tet
{

namespace classification
{

namespace classifiers
{

void SVM::train_classifier(const std::vector<Eigen::VectorXf*> & pos,
                           const std::vector<Eigen::VectorXf*> & neg,
                           const std::string weightsFile)
{
    Eigen::MatrixXi posAssignments(pos.size(), 1);
    posAssignments.setConstant(1);
    Eigen::MatrixXi negAssignments(neg.size(), 1);
    negAssignments.setConstant(-1);

    std::string posSVMLight = tet::classification::io::toSVMLight(pos, posAssignments);
    std::string negSVMLight = tet::classification::io::toSVMLight(neg, negAssignments);

    // write training data to file
    std::ofstream trainFile;
    trainFile.open(".tmpTrainSVM");
    trainFile << posSVMLight;
    trainFile << negSVMLight;
    trainFile.close();

    // call svm light
    std::cout << "[SVM::train | SVMLight] Starting..." << std::endl;
    // create arguments from c and epsilon
    std::stringstream epsStr;
    if (_eps >= 0)
    {
        epsStr << " -w " << _eps << " ";
    }
    std::stringstream cStr;
    if (_c >= 0)
    {
        cStr << " -c " << _c << " ";
    }

    std::stringstream command;
    command << SVM_LEARN << epsStr.str() << cStr.str() << util::kernelToSVMLightArgs(_kernel)
            << ".tmpTrainSVM " << _svmModelFile << " -a .trainResults.txt";
    std::cout << command.str() << std::endl;

    FILE * svmLightStream;
    std::cout << command.str().c_str() << std::endl;
    svmLightStream = popen(command.str().c_str(), "r");
    if (svmLightStream == NULL) {
        std::cerr << "[SVM::train] Error: Could not start SVMLight!" << std::endl;
        return;
    }

    std::ostringstream svmOutput;
    while (!feof(svmLightStream) && !ferror(svmLightStream))
    {
        char buf[128];
        int bytesRead = fread(buf, 1, 128, svmLightStream);
        svmOutput.write(buf, bytesRead);
    }
    std::cout << svmOutput.str() << std::endl;
    std::cout << "[SVM::train | SVMLight] Done. Output model written to " << _svmModelFile << "." << std::endl;

    pclose(svmLightStream);
    svmOutput.str("");
    command.str("");

    command << SVM2WEIGHTS << " " << _svmModelFile << " > " << weightsFile;

    svmLightStream = popen(command.str().c_str(), "r");
    if (svmLightStream == NULL) {
        std::cerr << "[SVM::train] Error: Could not start svm2weights!" << std::endl;
        return;
    }

    while (!feof(svmLightStream) && !ferror(svmLightStream))
    {
        char buf[128];
        int bytesRead = fread(buf, 1, 128, svmLightStream);
        svmOutput.write(buf, bytesRead);
    }

    pclose(svmLightStream);
}

void SVM::test_classifier(const std::vector<Eigen::VectorXf*> & test,
                          std::string resultFile)
{
    Eigen::MatrixXi zeros(test.size(), 1);
    zeros.setZero();

    std::string svmLight = tet::classification::io::toSVMLight(test, zeros);

    // write test data to file
    std::ofstream testFile;
    testFile.open(".tmpTestSVM");
    testFile << svmLight;
    testFile.close();

    // call svm light
    std::cout << "[SVM::test | SVMLight] Starting..." << std::endl;

    FILE * svmLightStream;
    
    std::stringstream command;
    command << SVM_CLASSIFY << " .tmpTestSVM " << _svmModelFile << " " << resultFile;
    std::cout << command.str().c_str() << std::endl;
    svmLightStream = popen(command.str().c_str(), "r");
    if (svmLightStream == NULL) {
        std::cerr << "[SVM::train] Error: Could not start SVMLight!" << std::endl;
        return;
    }

    std::ostringstream svmOutput;
    while (!feof(svmLightStream) && !ferror(svmLightStream))
    {
        char buf[128];
        int bytesRead = fread(buf, 1, 128, svmLightStream);
        svmOutput.write(buf, bytesRead);
    }
    std::cout << svmOutput.str() << std::endl;

    pclose(svmLightStream);

    std::cout << "[SVM::train | SVMLight] Done. Test results written to " << resultFile << "." << std::endl;
    
}

void SVM::test_classifier_existing(const std::vector<Eigen::VectorXf*> & test,
                                   std::string resultFile,
                                   std::string weightsFile)
{
    std::cout << "Testing with existing weights" << std::endl;

    Eigen::VectorXf weights(test[0]->rows());
    std::ifstream weightsFileStream(weightsFile);
    std::string line;

    float bias;
    if (weightsFileStream.is_open())
    {
        getline(weightsFileStream, line);
        bias = boost::lexical_cast<float>(line);
        std::cout << bias << std::endl;

        size_t i = 0;
        while (getline(weightsFileStream, line))
        {
            weights(i) = boost::lexical_cast<float>(line);
            ++i;
        }
    
        // write result data to file
        std::ofstream resultFileStream;
        resultFileStream.open(resultFile);
        for (size_t i = 0; i < test.size(); ++i)
        {
            float score = weights.transpose() * (*(test[i])) - bias;
            resultFileStream << score << std::endl;
        }

        resultFileStream.close();

    } else {
        std::cerr << "[SVM::test_classifier_existing] Could not open weights file" << std::endl;
        return;
    }
    
}

void SVM::train_regressor(const std::vector<Eigen::VectorXf*> & samples,
                          const std::vector<float> values,
                          const std::string weightsFile)
{
    Eigen::VectorXf eigValues = Eigen::VectorXf::Map(&values[0], values.size());
    std::string SVMLight = tet::classification::io::toSVMLight(samples, eigValues);

    // write training data to file
    std::ofstream trainFile;
    trainFile.open(".tmpTrainSVM");
    trainFile << SVMLight;
    trainFile.close();

    // call svm light
    std::cout << "[SVM::train_regressor | SVMLight] Starting..." << std::endl;
    // create arguments from c and epsilon
    std::stringstream epsStr;
    if (_eps >= 0)
    {
        epsStr << " -w " << _eps << " ";
    }
    std::stringstream cStr;
    if (_c >= 0)
    {
        cStr << " -c " << _c << " ";
    }

    // construct command
    std::stringstream command;
    command << SVM_LEARN << " -z r " << epsStr.str() << cStr.str() << util::kernelToSVMLightArgs(_kernel)
            << ".tmpTrainSVM " << _svmModelFile << " -a .trainResults.txt";
    std::cout << command.str() << std::endl;

    FILE * svmLightStream;
    svmLightStream = popen(command.str().c_str(), "r");
    if (svmLightStream == NULL) {
        std::cerr << "[SVM::train] Error: Could not start SVMLight!" << std::endl;
        return;
    }

    std::ostringstream svmOutput;
    while (!feof(svmLightStream) && !ferror(svmLightStream))
    {
        char buf[128];
        int bytesRead = fread(buf, 1, 128, svmLightStream);
        for (int i = 0; i < bytesRead; ++i)
        {
            std::cout << buf[i];
        }
        std::cout << std::flush;
        svmOutput.write(buf, bytesRead);
    }
    std::cout << svmOutput.str() << std::endl;
    std::cout << "[SVM::train_regressor | SVMLight] Done. Output model written to " << _svmModelFile << "." << std::endl;

    pclose(svmLightStream);
    svmOutput.str("");
    command.str("");

    command << SVM2WEIGHTS << " " << _svmModelFile << " > " << weightsFile;

    svmLightStream = popen(command.str().c_str(), "r");
    if (svmLightStream == NULL) {
        std::cerr << "[SVM::train] Error: Could not start svm2weights!" << std::endl;
        return;
    }

    while (!feof(svmLightStream) && !ferror(svmLightStream))
    {
        char buf[128];
        int bytesRead = fread(buf, 1, 128, svmLightStream);
        svmOutput.write(buf, bytesRead);
    }

    pclose(svmLightStream);
}

void SVM::test_regressor(const std::vector<Eigen::VectorXf*> & test,
                         std::string resultFile)
{
    Eigen::MatrixXi zeros(test.size(), 1);
    zeros.setZero();

    std::string svmLight = tet::classification::io::toSVMLight(test, zeros);

    // write test data to file
    std::ofstream testFile;
    testFile.open(".tmpTestSVM");
    testFile << svmLight;
    testFile.close();

    // call svm light
    std::cout << "[SVM::test_regressor | SVMLight] Starting..." << std::endl;

    FILE * svmLightStream;
    
    std::stringstream command;
    command << SVM_CLASSIFY << " .tmpTestSVM " << _svmModelFile << " " << resultFile;
    std::cout << command.str().c_str() << std::endl;
    svmLightStream = popen(command.str().c_str(), "r");
    if (svmLightStream == NULL) {
        std::cerr << "[SVM::test_regressor] Error: Could not start SVMLight!" << std::endl;
        return;
    }

    std::ostringstream svmOutput;
    while (!feof(svmLightStream) && !ferror(svmLightStream))
    {
        char buf[128];
        int bytesRead = fread(buf, 1, 128, svmLightStream);
        svmOutput.write(buf, bytesRead);
    }
    std::cout << svmOutput.str() << std::endl;

    pclose(svmLightStream);

    std::cout << "[SVM::test_regressor | SVMLight] Done. Test results written to " << resultFile << "." << std::endl;
    
}

void SVM::test_regressor_existing(const std::vector<Eigen::VectorXf*> & test,
                                  std::string resultFile,
                                  std::string weightsFile)
{
    std::cout << "Testing with existing weights" << std::endl;

    Eigen::VectorXf weights(test[0]->rows());
    std::ifstream weightsFileStream(weightsFile);
    std::string line;
    float bias;
    if (weightsFileStream.is_open())
    {
        getline(weightsFileStream, line);
        bias = boost::lexical_cast<float>(line);
        std::cout << bias << std::endl;

        size_t i = 0;
        while (getline(weightsFileStream, line))
        {
            weights(i) = boost::lexical_cast<float>(line);
            ++i;
        }
        // write result data to file
        std::ofstream resultFileStream;
        resultFileStream.open(resultFile);
        for (size_t i = 0; i < test.size(); ++i)
        {
            float score = weights.transpose() * (*(test[i])) - bias;
            resultFileStream << score << std::endl;
        }
        resultFileStream.close();
    } else {
        std::cerr << "[SVM::test_regressor_existing] Could not open weights file" << std::endl;
    }

}

} // namespace classifiers

} // namespace classification

} // namespace tet
