#include <boost/functional/hash.hpp>

#include "tet/descriptors/cnndescriptor.h"

#include <vector>

namespace tet
{

namespace classification
{

namespace descriptors
{

void CNNDescriptor::_setup()
{
    caffe::Caffe::set_mode(caffe::Caffe::CPU);
    caffe::Caffe::SetDevice(0);
}

void CNNDescriptor::setNetProto(std::string netProtoFile)
{
    _deployFile = netProtoFile;
    _net = new caffe::Net<float>(netProtoFile, caffe::Phase::TEST);
}

bool CNNDescriptor::setTrainedModel(std::string modelFile)
{
    if (_net == nullptr)
    {
        std::cerr << "Net not initialized yet. Call setNetProto first." << std::endl;
        return false;
    }

    _net->CopyTrainedLayersFrom(modelFile.c_str());
    _modelFile = modelFile;
    return true;
}

cv::Mat CNNDescriptor::getVisualization(const FeatureType &features) const
{
    return cv::Mat();
}

Eigen::MatrixXf CNNDescriptor::getDescriptors(const InputType &img, VisType *vis) const
{
    // resize image to mean file
    cv::Mat rszimg;
    cv::resize(img, rszimg, cv::Size(256, 256));

    float loss = 0.0;
    std::vector<cv::Mat> vecData = {rszimg};
    std::vector<int>     vecLbl  = {0};
    boost::shared_ptr<caffe::MemoryDataLayer<float> > md_layer =
        boost::dynamic_pointer_cast<caffe::MemoryDataLayer<float> >(_net->layers()[0]);
    if (!md_layer) {
        std::cout << "The first layer is not a MemoryDataLayer!" << std::endl;
    }

    md_layer->AddMatVector(vecData, vecLbl);
    std::vector<caffe::Blob<float>* > results = _net->ForwardPrefilled(&loss);

    boost::shared_ptr<caffe::Blob<float> > out = _net->blob_by_name(_outLayer);

    Eigen::MatrixXf res(out->count(), 1);
    for (int i = 0; i < out->count(); ++i)
    {
        res(i, 0) = out->cpu_data()[i];
    }
    res.normalize();

    return res;
}

tinyxml2::XMLElement* CNNDescriptor::serialize(tinyxml2::XMLElement* parent,
                                               tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *root = doc->NewElement("descriptor");
    root->SetAttribute("type", "cnn");
    parent->LinkEndChild(root);

    tinyxml2::XMLElement *deployProto = doc->NewElement("deploy");
    deployProto->SetText(_deployFile.c_str());
    root->LinkEndChild(deployProto);

    tinyxml2::XMLElement *modelFile = doc->NewElement("modelFile");
    modelFile->SetText(_modelFile.c_str());
    root->LinkEndChild(modelFile);

    tinyxml2::XMLElement *outLayer = doc->NewElement("outLayer");
    outLayer->SetText(_outLayer.c_str());
    root->LinkEndChild(outLayer);

    return root;
}

void CNNDescriptor::deserialize(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *deployProto = root->FirstChildElement("deploy"))
    {
        std::string deployFile = deployProto->GetText();
        this->setNetProto(deployFile);
    }

    if (const tinyxml2::XMLElement *modelFileTag = root->FirstChildElement("modelFile"))
    {
        std::string modelFile = modelFileTag->GetText();
        this->setTrainedModel(modelFile);
    }

    if (const tinyxml2::XMLElement *outLayerTag = root->FirstChildElement("outLayer"))
    {
        _outLayer = outLayerTag->GetText();
    }

}

std::size_t hash_value(const CNNDescriptor & desc)
{
    std::size_t seed = 0;
    boost::hash_combine(seed, desc._deployFile);
    boost::hash_combine(seed, desc._modelFile);
    boost::hash_combine(seed, desc._outLayer);

    return seed;
}

} // namespace descriptors

} // namespace classification

} // namespace tet

namespace std
{

    size_t hash<tet::classification::descriptors::CNNDescriptor>::operator()
    (const tet::classification::descriptors::CNNDescriptor & desc) const
    {
        return 0;
    }

} // namespace std

