#include "tet/descriptors/hogdescriptor.h"
#include "opencv2/core/eigen.hpp"

#include <fstream>

namespace tet
{

namespace classification
{

namespace descriptors
{

cv::Mat HOGDescriptor::getVisualization(const FeatureType &features) const
{
    vl_size hogWidth = vl_hog_get_width(_vlHog);
    vl_size hogHeight = vl_hog_get_height(_vlHog);
    vl_size hogDimension = vl_hog_get_dimension(_vlHog);
    float *hogArray = (float*)vl_malloc(hogWidth*hogHeight*hogDimension*sizeof(float));

    std::cout << hogWidth << ", " << hogHeight << std::endl;
    std::cout << hogDimension << std::endl;
    Eigen::Map<Eigen::MatrixXf>(hogArray, hogWidth*hogHeight, hogDimension) = features;

    vl_size glyphSize = vl_hog_get_glyph_size(_vlHog);
    int renImageHeight = glyphSize * hogHeight;
    int renImageWidth  = glyphSize * hogWidth;
    float *renImage = (float*)vl_malloc(renImageHeight * renImageWidth * sizeof(float));
    vl_hog_render(_vlHog, renImage, hogArray, hogWidth, hogHeight);
    cv::Mat vis(renImageHeight, renImageWidth, CV_32FC1, renImage);
    vis.convertTo(vis, CV_8UC1, 255.0);

    return vis;
}

Eigen::MatrixXf HOGDescriptor::getDescriptors(const InputType &img, VisType* vis) const
{

    const float *imgPtr;
    cv::Mat concat;
    if (img.channels() > 1)
    {
        std::vector<cv::Mat> channels;
        for (int i = 0; i < img.channels(); ++i)
        {
            channels.push_back(cv::Mat(img.rows, img.cols, CV_32FC1));
        }
        cv::split(img, channels);

        cv::vconcat(channels, concat);

        imgPtr = concat.ptr<float>(0);
    } else {
        imgPtr = img.ptr<float>(0);
    }
    float* hogArray;

    vl_hog_put_image(_vlHog, imgPtr, img.cols, img.rows, img.channels(), _cellSize);
    vl_size hogWidth = vl_hog_get_width(_vlHog);
    vl_size hogHeight = vl_hog_get_height(_vlHog);
    vl_size hogDimension = vl_hog_get_dimension(_vlHog);

    hogArray = (float*)vl_malloc(hogWidth*hogHeight*hogDimension*sizeof(float));
    vl_hog_extract(_vlHog, (float*)hogArray);

    if (vis)
    {
        vl_size glyphSize = vl_hog_get_glyph_size(_vlHog);
        int renImageHeight = glyphSize * hogHeight;
        int renImageWidth  = glyphSize * hogWidth;
        float *renImage = (float*)vl_malloc(renImageHeight * renImageWidth * sizeof(float));
        vl_hog_render(_vlHog, renImage, hogArray, hogWidth, hogHeight);
        *vis = cv::Mat(renImageHeight, renImageWidth, CV_32FC1, renImage);
        vis->convertTo(*vis, CV_8UC1, 255.0);
    }

    Eigen::MatrixXf descriptors;
    descriptors = _includeLocation ? Eigen::MatrixXf(hogWidth * hogHeight, hogDimension + 2)
                                   : Eigen::MatrixXf(hogWidth * hogHeight, hogDimension);

    if (_removeZeroGradient)
    {
        Eigen::MatrixXf tmp(hogWidth*hogHeight, hogDimension);
        tmp = Eigen::Map<Eigen::MatrixXf>(hogArray, hogWidth*hogHeight, hogDimension);

        int descIdx = 0;
        for (int i = 0; i < tmp.rows(); ++i)
        {
            if (tmp.row(i).norm() > 0.01)
            {
                int x = i % hogWidth;
                int y = i / hogWidth;
                
                descriptors.block(descIdx, 0, 1, hogDimension) = tmp.row(i);
                if (_includeLocation)
                {
                    descriptors(descIdx, hogDimension) = static_cast<float>(x) / hogWidth;
                    descriptors(descIdx, hogDimension + 1) = static_cast<float>(y) / hogHeight;
                }
                ++descIdx;
            } 
        }
        descriptors.conservativeResize(descIdx, (_includeLocation ? hogDimension + 2 : hogDimension));
    
        vl_free(hogArray);
        return descriptors;

    } else {
        Eigen::MatrixXf tmp(hogWidth*hogHeight, hogDimension);
        tmp = Eigen::Map<Eigen::MatrixXf>(hogArray, hogWidth*hogHeight, hogDimension);

        for (int i = 0; i < tmp.rows(); ++i)
        {
            int x = i % hogWidth;
            int y = i / hogWidth;
            
            descriptors.block(i, 0, 1, hogDimension) = tmp.row(i);

            if (_includeLocation)
            {
                descriptors(i, hogDimension) = static_cast<float>(x) / hogWidth;
                descriptors(i, hogDimension + 1) = static_cast<float>(y) / hogHeight;
            }
        }

        vl_free(hogArray);
        return descriptors;
    }
}

tinyxml2::XMLElement* HOGDescriptor::serialize(tinyxml2::XMLElement* parent,
                                               tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *root = doc->NewElement("descriptor");
    root->SetAttribute("type", "hog");
    parent->LinkEndChild(root);
    
    tinyxml2::XMLElement *cellSize = doc->NewElement("cellSize");
    cellSize->SetText(_cellSize);
    root->LinkEndChild(cellSize);

    tinyxml2::XMLElement *bins = doc->NewElement("bins");
    bins->SetText(_nbins);
    root->LinkEndChild(bins);

    tinyxml2::XMLElement *removeZeroGradient = doc->NewElement("removeZeroGradient");
    removeZeroGradient->SetText((_removeZeroGradient ? "1" : "0"));
    root->LinkEndChild(removeZeroGradient);

    tinyxml2::XMLElement *includeLocation = doc->NewElement("includeLocation");
    includeLocation->SetText((_includeLocation ? "1" : "0"));
    root->LinkEndChild(includeLocation);

    return root;

}

void HOGDescriptor::deserialize(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *cellSize = root->FirstChildElement("cellSize"))
    {
        cellSize->QueryIntText(&_cellSize);
    }

    if (const tinyxml2::XMLElement *bins = root->FirstChildElement("bins"))
    {
        bins->QueryIntText(&_nbins);
    }

    if (const tinyxml2::XMLElement *removeZeroGradient = root->FirstChildElement("removeZeroGradient"))
    {
        int tmp;
        removeZeroGradient->QueryIntText(&tmp);
        _removeZeroGradient = tmp > 0;
    }

    if (const tinyxml2::XMLElement *includeLocation = root->FirstChildElement("includeLocation"))
    {
        int tmp;
        includeLocation->QueryIntText(&tmp);
        _includeLocation = tmp > 0;
    }
}

std::size_t hash_value(const HOGDescriptor & desc)
{
    std::size_t seed = 0;
    boost::hash_combine(seed, desc._nbins);
    boost::hash_combine(seed, desc._cellSize);
    boost::hash_combine(seed, desc._removeZeroGradient);
    boost::hash_combine(seed, desc._includeLocation);

    return seed;
}

} // namespace descriptors

} // namespace classification

} // namespace tet

namespace std
{

    size_t hash<tet::classification::descriptors::HOGDescriptor>::operator()
    (const tet::classification::descriptors::HOGDescriptor & desc) const
    {
        size_t nbins = std::hash<int>()(desc._nbins);
        size_t cellSize = std::hash<int>()(desc._cellSize);
        size_t removeZeroGradient = std::hash<bool>()(desc._removeZeroGradient);
        size_t includeLocation = std::hash<bool>()(desc._includeLocation);

        return nbins ^ (cellSize << 1) ^ (removeZeroGradient << 2) ^
              (includeLocation << 3);
    }

} // namespace std

