#include "tet/descriptors/hogdescriptorspb.h"
#include "opencv2/core/eigen.hpp"

#include <fstream>

namespace tet
{

namespace classification
{

namespace descriptors
{

cv::Mat HOGDescriptorSPB::getVisualization(const FeatureType &features) const
{
    std::vector<cv::Mat> visualizations;

    int visRows = 0;
    int visCols = 0;

    for (size_t i = 0; i < features.size(); ++i)
    {
        vl_size hogWidth = vl_hog_get_width(_vlHog);
        vl_size hogHeight = vl_hog_get_height(_vlHog);
        vl_size hogDimension = vl_hog_get_dimension(_vlHog);
        float *hogArray = (float*)vl_malloc(hogWidth*hogHeight*hogDimension*sizeof(float));

        std::cout << hogWidth << ", " << hogHeight << std::endl;
        std::cout << hogDimension << std::endl;
        Eigen::Map<Eigen::MatrixXf>(hogArray, hogWidth*hogHeight, hogDimension) = features[i];

        vl_size glyphSize = vl_hog_get_glyph_size(_vlHog);
        int renImageHeight = glyphSize * hogHeight;
        int renImageWidth  = glyphSize * hogWidth;
        float *renImage = (float*)vl_malloc(renImageHeight * renImageWidth * sizeof(float));
        vl_hog_render(_vlHog, renImage, hogArray, hogWidth, hogHeight);
        cv::Mat vis(renImageHeight, renImageWidth, CV_32FC1, renImage);
        vis.convertTo(vis, CV_8UC1, 255.0);
        std::cout << renImageHeight << ", " << renImageWidth << std::endl;
        cv::namedWindow("img");
        cv::imshow("img", vis);
        cv::waitKey(0);
        visRows += vis.rows;
        visCols += vis.cols;
        visualizations.push_back(vis);
    }

    cv::Mat visFull(visRows, visCols, CV_8UC1);
    int visRowsPerBlock = visRows / _ny;
    int visColsPerBlock = visCols / _nx;

    for (int y = 0; y < _ny; ++y)
    {
        for (int x = 0; x < _nx; ++x)
        {
            cv::Mat curBlock = visFull(cv::Rect(x * visColsPerBlock, y * visRowsPerBlock, visColsPerBlock, visRowsPerBlock));
            curBlock = visualizations[y * _nx + x];
        }
    }

    return visFull;
}

std::vector<Eigen::MatrixXf> HOGDescriptorSPB::getDescriptors(const InputType &img, VisType* vis) const
{
    // split up img into _ny x _ny blocks
    int rowsPerBlock = img.rows / _ny;
    int colsPerBlock = img.cols / _nx;

    std::vector<Eigen::MatrixXf> descriptors;
    std::vector<cv::Mat> visualizations;

    int visRows = 0;
    int visCols = 0;

    for (int y = 0; y < _ny; ++y)
    {
        for (int x = 0; x < _nx; ++x)
        {
            cv::Mat curBlock = cv::Mat(img, cv::Rect(x * colsPerBlock, y * rowsPerBlock, colsPerBlock, rowsPerBlock));

            if (vis)
            {
                cv::Mat curVis;
                descriptors.push_back(this->getLocalDescriptors(curBlock, &curVis));
                visualizations.push_back(curVis);
            } else {
                descriptors.push_back(this->getLocalDescriptors(curBlock, nullptr));
            }
        }
    }

    if (vis)
    {
        visRows = visualizations[0].rows * _ny;
        visCols = visualizations[0].cols * _nx;

        int visRowsPerBlock = visRows / _ny;
        int visColsPerBlock = visCols / _nx;

        *vis = cv::Mat(visRows, visCols, CV_8UC1);
        std::cout << "---" << std::endl;
        std::cout << visRowsPerBlock << ", " << visColsPerBlock << std::endl;
        std::cout << "---" << std::endl;
        for (int y = 0; y < _ny; ++y)
        {
            for (int x = 0; x < _nx; ++x)
            {
                /*cv::Mat block(*vis,
                cv::Mat curBlock = (*vis)(cv::Rect(x * visColsPerBlock, y * visRowsPerBlock, visColsPerBlock, visRowsPerBlock));
                std::cout << x*visColsPerBlock << ", " << y*visRowsPerBlock << std::endl;*/

                //std::cout << vis->rows << ", " << vis->cols << std::endl;
                //std::cout << visualizations[y * _nx + x].rows << ", " << visualizations[y * _nx + x].cols << std::endl;
                std::cout << visualizations[y * _nx + x].rows << ", " << visualizations[y * _nx + x].cols << std::endl;
                visualizations[y * _nx + x].copyTo((*vis)(cv::Rect(x * visColsPerBlock, y * visRowsPerBlock, visColsPerBlock, visRowsPerBlock)));
                //curBlock = visualizations[y * _nx + x];
                //curBlock = cv::Scalar(255);
                //vis(cv::Rect(x * visColsPerBlock, y * visRowsPerBlock, visColsPerBlock, visRowsPerBlock)) = visualizations[y * _nx + x];
            }
        }
    }

    return descriptors;

}

Eigen::MatrixXf HOGDescriptorSPB::getLocalDescriptors(const InputType &img, VisType* vis) const
{
    const float *imgPtr;
    cv::Mat concat;
    if (img.channels() > 1)
    {
        std::vector<cv::Mat> channels;
        for (int i = 0; i < img.channels(); ++i)
        {
            channels.push_back(cv::Mat(img.rows, img.cols, CV_32FC1));
        }
        cv::split(img, channels);

        cv::vconcat(channels, concat);

        imgPtr = concat.ptr<float>(0);
    } else {
        imgPtr = img.ptr<float>(0);
    }
    float* hogArray;

    vl_hog_put_image(_vlHog, imgPtr, img.cols, img.rows, img.channels(), _cellSize);
    vl_size hogWidth = vl_hog_get_width(_vlHog);
    vl_size hogHeight = vl_hog_get_height(_vlHog);
    vl_size hogDimension = vl_hog_get_dimension(_vlHog);

    hogArray = (float*)vl_malloc(hogWidth*hogHeight*hogDimension*sizeof(float));
    vl_hog_extract(_vlHog, (float*)hogArray);

    /*for (int i = 0; i < hogWidth; ++i)
    {
        for (int j = 0; j < hogHeight; ++j)
        {
            for (int k = 0; k < hogDimension; ++k)
            {
                std::cout << hogArray[i * hogHeight * hogDimension + j * hogHeight + k] << std::endl;
            }
        }
    }*/

    if (vis)
    {
        vl_size glyphSize = vl_hog_get_glyph_size(_vlHog);
        int renImageHeight = glyphSize * hogHeight;
        int renImageWidth  = glyphSize * hogWidth;
        float *renImage = (float*)vl_malloc(renImageHeight * renImageWidth * sizeof(float));
        vl_hog_render(_vlHog, renImage, hogArray, hogWidth, hogHeight);
        *vis = cv::Mat(renImageHeight, renImageWidth, CV_32FC1, renImage);
        vis->convertTo(*vis, CV_8UC1, 255.0);
    }

    if (_removeZeroGradient)
    {
        Eigen::MatrixXf tmp(hogWidth*hogHeight, hogDimension);
        tmp = Eigen::Map<Eigen::MatrixXf>(hogArray, hogWidth*hogHeight, hogDimension);

        Eigen::MatrixXf descriptors(hogWidth*hogHeight, hogDimension + 2); // 2 is for X, Y

        int descIdx = 0;
        for (int i = 0; i < tmp.rows(); ++i)
        {
            if (tmp.row(i).norm() > 0.01)
            {
                int x = i % hogWidth;
                int y = i / hogWidth;
                
                descriptors.block(descIdx, 0, 1, hogDimension) = tmp.row(i);
                descriptors(descIdx, hogDimension) = static_cast<float>(x) / hogWidth;
                descriptors(descIdx, hogDimension + 1) = static_cast<float>(y) / hogHeight;
                ++descIdx;
            } 
        }
        descriptors.conservativeResize(descIdx, hogDimension + 2);

        return descriptors;

    } else {
        Eigen::MatrixXf tmp(hogWidth*hogHeight, hogDimension);
        tmp = Eigen::Map<Eigen::MatrixXf>(hogArray, hogWidth*hogHeight, hogDimension);

        Eigen::MatrixXf descriptors(hogWidth*hogHeight, hogDimension + 2); // 2 is for X, Y

        for (int i = 0; i < tmp.rows(); ++i)
        {
            int x = i % hogWidth;
            int y = i / hogWidth;
            
            descriptors.block(i, 0, 1, hogDimension) = tmp.row(i);
            descriptors(i, hogDimension) = static_cast<float>(x) / hogWidth;
            descriptors(i, hogDimension + 1) = static_cast<float>(y) / hogHeight;
        }

        return descriptors;
    }
}

tinyxml2::XMLElement* HOGDescriptorSPB::serialize(tinyxml2::XMLElement* parent,
                                                  tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *root = doc->NewElement("descriptor");
    root->SetAttribute("type", "hogspb");
    parent->LinkEndChild(root);
    
    tinyxml2::XMLElement *cellSize = doc->NewElement("cellSize");
    cellSize->SetText(_cellSize);
    root->LinkEndChild(cellSize);

    tinyxml2::XMLElement *bins = doc->NewElement("bins");
    bins->SetText(_nbins);
    root->LinkEndChild(bins);

    tinyxml2::XMLElement *removeZeroGradient = doc->NewElement("removeZeroGradient");
    removeZeroGradient->SetText((_removeZeroGradient ? "1" : "0"));
    root->LinkEndChild(removeZeroGradient);

    tinyxml2::XMLElement *gridSize = doc->NewElement("gridSize");
    tinyxml2::XMLElement *nx = doc->NewElement("nx");
    nx->SetText(_nx);
    tinyxml2::XMLElement *ny = doc->NewElement("ny");
    ny->SetText(_ny);
    gridSize->LinkEndChild(nx);
    gridSize->LinkEndChild(ny);
    root->LinkEndChild(gridSize);

    return root;

}

void HOGDescriptorSPB::deserialize(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *cellSize = root->FirstChildElement("cellSize"))
    {
        cellSize->QueryIntText(&_cellSize);
    }

    if (const tinyxml2::XMLElement *bins = root->FirstChildElement("bins"))
    {
        bins->QueryIntText(&_nbins);
    }

    if (const tinyxml2::XMLElement *removeZeroGradient = root->FirstChildElement("removeZeroGradient"))
    {
        int tmp;
        removeZeroGradient->QueryIntText(&tmp);
        _removeZeroGradient = tmp > 0;
    }

    if (const tinyxml2::XMLElement *gridSize = root->FirstChildElement("gridSize"))
    {
        if (const tinyxml2::XMLElement *nx = gridSize->FirstChildElement("nx"))
        {
            nx->QueryIntText(&_nx);
        }

        if (const tinyxml2::XMLElement *ny = gridSize->FirstChildElement("ny"))
        {
            ny->QueryIntText(&_ny);
        }
    }

}

} // namespace descriptors

} // namespace classification

} // namespace tet

namespace std
{

    size_t hash<tet::classification::descriptors::HOGDescriptorSPB>::operator()
    (const tet::classification::descriptors::HOGDescriptorSPB & desc) const
    {
        size_t nbins = std::hash<int>()(desc._nbins);
        size_t cellSize = std::hash<int>()(desc._cellSize);
        size_t removeZeroGradient = std::hash<bool>()(desc._removeZeroGradient);

        size_t nx = std::hash<int>()(desc._nx);
        size_t ny = std::hash<int>()(desc._ny);

        return nbins ^ (cellSize << 1) ^ (removeZeroGradient << 2)
             ^ (nx << 3) ^ (ny << 4);
    }

} // namespace std
