#include <cmath>

#include "boost/lexical_cast.hpp"

#include "eigen3/Eigen/Core"
#include "eigen3/Eigen/Eigenvalues"
#include "opencv2/core/eigen.hpp"

#include "tet/encodings/kcencoding.h"

#include "tet/util/stringutils.h"

#define GAMMA 200

namespace tet
{

namespace classification
{

namespace encodings
{

template <class _VecType>
float KCDistance(_VecType x, _VecType m)
{
    return std::exp(-(GAMMA / 2)*(x - m).squaredNorm());
}

void KCEncoding::setCenters(SampleType centers)
{
    if (_tree)
    {
        delete _tree;
    }

    _centers = centers;
    _tree = new tet::classification::util::ANNTree<Scalar>(_centers);

    _computed = true;
}

KCEncoding::FeatureType KCEncoding::computeFeatures(SampleType samples) const
{
    FeatureType kc(_k, 1);
    kc.setZero();

    for (int sampleIdx = 0; sampleIdx < samples.rows(); ++sampleIdx)
    {
        std::vector<size_t> nn = _tree->findNN(samples.row(sampleIdx), 4);

        float responseSum = 0.f;
        for (int nnIdx = 0; nnIdx < 4; ++nnIdx)
        {
            responseSum += KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[nnIdx]));
        }

        for (int nnIdx = 0; nnIdx < 4; ++nnIdx)
        {
            if (_maxPooling)
            {
                float response = KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[nnIdx]));
                kc(nn[nnIdx]) = ((response / responseSum) > kc(nn[nnIdx])) ? (response / responseSum) : kc(nn[nnIdx]);
            } else {
                kc(nn[nnIdx]) += KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[nnIdx])) / responseSum;
            }
        }
    }

    if (samples.rows() > 0 && !_maxPooling)
    {
        kc /= samples.rows();
    }

    return kc;
}

tinyxml2::XMLElement* KCEncoding::serializeSpec(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *k = doc->NewElement("k");
    k->SetText(_k);
    parent->LinkEndChild(k);

    if (_computed)
    {
        tinyxml2::XMLElement *centers = doc->NewElement("centers");
        parent->LinkEndChild(centers);

        for (int i = 0; i < _centers.rows(); ++i)
        {
            tinyxml2::XMLElement *center = doc->NewElement("center");
            centers->LinkEndChild(center);

            std::stringstream row;
            for (int j = 0; j < _centers.row(i).cols() - 1; ++j)
            {
                row << _centers.row(i)(j) << " ";
            }
            row << _centers.row(i)(_centers.row(i).cols() - 1);
            center->SetText(row.str().c_str());
        }
    }

    tinyxml2::XMLElement *maxPooling = doc->NewElement("maxPooling");
    maxPooling->SetText((_maxPooling) ? "1" : "0");
    parent->LinkEndChild(maxPooling);

    tinyxml2::XMLElement *pca = doc->NewElement("pca");
    pca->SetText(_pcaDim);
    parent->LinkEndChild(pca);

    return parent;
}

void KCEncoding::deserializeSpec(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *k = root->FirstChildElement("k"))
    {
        k->QueryIntText(&_k);
    }

    if (const tinyxml2::XMLElement *samples = root->FirstChildElement("samples"))
    {
        if (const tinyxml2::XMLElement *set = samples->FirstChildElement("set"))
        {
            _samples.deserialize(set);
        } else if (const tinyxml2::XMLElement *file = samples->FirstChildElement("file"))
        {
            _samples.deserializeFromFile(std::string(file->GetText()));
        }
    }

    if (const tinyxml2::XMLElement *centers = root->FirstChildElement("centers"))
    {
        _computed = true;

        const tinyxml2::XMLElement *center = centers->FirstChildElement("center");
        std::vector<std::string> elems = stringSplit(std::string(center->GetText()), ' ');
        _centers = SampleType(_k, elems.size());
        std::cout << elems.size() << std::endl;
        int row = 0;
        for (size_t i = 0; i < elems.size(); ++i)
        {
            _centers(row, i) = boost::lexical_cast<Scalar>(elems[i]);
        }

        row = 1;
        for (center = center->NextSiblingElement("center");
             center != NULL;
             center = center->NextSiblingElement("center"))
        {
            elems = stringSplit(std::string(center->GetText()), ' ');
            for (int i = 0; i < _centers.cols(); ++i)
            {
                _centers(row, i) = boost::lexical_cast<Scalar>(elems[i]);
            }
            row += 1;
        }

        _tree = new tet::classification::util::ANNTree<Scalar>(_centers);
    }

    if (const tinyxml2::XMLElement *maxPooling = root->FirstChildElement("maxPooling"))
    {
        int tmp;
        maxPooling->QueryIntText(&tmp);
        _maxPooling = tmp > 0;
    }

    if (const tinyxml2::XMLElement *pca = root->FirstChildElement("pca"))
    {
        int tmp;
        pca->QueryIntText(&tmp);
        if (tmp > 0)
        {
            _usePCA = true;
            _pcaDim = tmp;
        }
    }
}

} // namespace encodings

} // namespace classification

} // namespace tet
