#include <cmath>

#include "boost/lexical_cast.hpp"

#include "eigen3/Eigen/Core"
#include "eigen3/Eigen/Eigenvalues"
#include "opencv2/core/eigen.hpp"

#include "tet/encodings/kcencodingspb.h"

#include "tet/util/stringutils.h"

namespace tet
{

namespace classification
{

namespace encodings
{

template <class _VecType>
float KCDistance(_VecType x, _VecType m)
{
    return std::exp(-(x - m).squaredNorm());
}

/*void KCEncodingSPB::init(InitStruct initStruct)
{
    cv::Mat samplesCV, labels, centers;

    // convert samples to OpenCV format
    cv::eigen2cv(*(initStruct.samples), samplesCV);
    samplesCV.convertTo(samplesCV, CV_32F);

    // compute kmeans
    kmeans(samplesCV, _k, labels, cv::TermCriteria(cv::TermCriteria::EPS + cv::TermCriteria::COUNT, 10, 0.001),
           3, cv::KMEANS_PP_CENTERS, centers);

    // convert back to Eigen
    cv2eigen(centers, _centers);

    //_tree = new tet::classification::util::FLANNTree<Scalar>(_centers);
    _tree = new tet::classification::util::ANNTree<Scalar>(_centers);

    _computed = true;

}*/

KCEncodingSPB::FeatureType KCEncodingSPB::computeFeatures(SampleType samples) const
{
    FeatureType kc(_k * samples.size(), 1);

    for (size_t i = 0; i < samples.size(); ++i)
    {
        kc.block(i * _k, 0, _k, 1) = this->computeFeaturesSingle(samples[i]);
    }
    
    return kc;
}

KCEncodingSPB::FeatureType KCEncodingSPB::computeFeaturesSingle(Eigen::Matrix<Scalar, -1, -1> samples) const
{
    FeatureType kc(_k, 1);
    kc.setZero();

    //std::vector<std::vector<size_t> > nn = _tree->findNN(samples, 4);

    for (int sampleIdx = 0; sampleIdx < samples.rows(); ++sampleIdx)
    {
        //std::vector<std::vector<size_t> > nn = _tree->findNN(samples.row(sampleIdx), 4);
        std::vector<size_t> nn = _tree->findNN(samples.row(sampleIdx), 4);
        for (int nnIdx = 0; nnIdx < 4; ++nnIdx)
        {
            if (_maxPooling)
            {
                float response = KCDistance<Eigen::MatrixXf>(samples.row(sampleIdx), _centers.row(nn[nnIdx]));
                //float response = KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[0][nnIdx]));
                //kc(nn[0][nnIdx]) = (response > kc(nn[0][nnIdx])) ? response : kc(nn[0][nnIdx]);
                kc(nn[nnIdx]) = (response > kc(nn[nnIdx])) ? response : kc(nn[nnIdx]);
                //std::cout << nn[nnIdx] << std::endl;
                //float response = KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[sampleIdx][nnIdx]));
                //kc(nn[sampleIdx][nnIdx]) = (response > kc(nn[sampleIdx][nnIdx])) ? response : kc(nn[sampleIdx][nnIdx]);
            } else {
                kc(nn[nnIdx]) += KCDistance<Eigen::MatrixXf>(samples.row(sampleIdx), _centers.row(nn[nnIdx]));
                //kc(nn[0][nnIdx]) += KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[0][nnIdx]));
                //kc(nn[sampleIdx][nnIdx]) += KCDistance<SampleType>(samples.row(sampleIdx), _centers.row(nn[sampleIdx][nnIdx]));
            }
        }
    }

    if (samples.rows() > 0 && !_maxPooling)
    {
        kc = kc.normalized();
    }

    return kc;
}

tinyxml2::XMLElement* KCEncodingSPB::serializeSpec(tinyxml2::XMLElement* parent, tinyxml2::XMLDocument* doc)
{
    tinyxml2::XMLElement *k = doc->NewElement("k");
    k->SetText(_k);
    parent->LinkEndChild(k);

    if (_computed)
    {
        tinyxml2::XMLElement *centers = doc->NewElement("centers");
        parent->LinkEndChild(centers);

        for (int i = 0; i < _centers.rows(); ++i)
        {
            tinyxml2::XMLElement *center = doc->NewElement("center");
            centers->LinkEndChild(center);

            std::stringstream row;
            for (int j = 0; j < _centers.row(i).cols() - 1; ++j)
            {
                row << _centers.row(i)(j) << " ";
            }
            row << _centers.row(i)(_centers.row(i).cols() - 1);
            center->SetText(row.str().c_str());
        }
    }

    tinyxml2::XMLElement *maxPooling = doc->NewElement("maxPooling");
    maxPooling->SetText((_maxPooling) ? "1" : "0");
    parent->LinkEndChild(maxPooling);

    return parent;
}

void KCEncodingSPB::deserializeSpec(const tinyxml2::XMLElement* root)
{
    if (const tinyxml2::XMLElement *k = root->FirstChildElement("k"))
    {
        k->QueryIntText(&_k);
    }

    if (const tinyxml2::XMLElement *centers = root->FirstChildElement("centers"))
    {
        _computed = true;

        const tinyxml2::XMLElement *center = centers->FirstChildElement("center");
        std::vector<std::string> elems = stringSplit(std::string(center->GetText()), ' ');
        _centers = Eigen::MatrixXf(_k, elems.size());
        std::cout << elems.size() << std::endl;
        int row = 0;
        for (size_t i = 0; i < elems.size(); ++i)
        {
            _centers(row, i) = boost::lexical_cast<Scalar>(elems[i]);
        }

        row = 1;
        for (center = center->NextSiblingElement("center");
             center != NULL;
             center = center->NextSiblingElement("center"))
        {
            elems = stringSplit(std::string(center->GetText()), ' ');
            for (int i = 0; i < _centers.cols(); ++i)
            {
                _centers(row, i) = boost::lexical_cast<Scalar>(elems[i]);
            }
            row += 1;
        }

        //_tree = new tet::classification::util::FLANNTree<Scalar>(_centers);
        _tree = new tet::classification::util::ANNTree<Scalar>(_centers);
    }

    if (const tinyxml2::XMLElement *maxPooling = root->FirstChildElement("maxPooling"))
    {
        int tmp;
        maxPooling->QueryIntText(&tmp);
        _maxPooling = tmp > 0;
    }
}

} // namespace encodings

} // namespace classification

} // namespace tet
