#include <iostream>
#include <iomanip>
#include <string>

#include "boost/lexical_cast.hpp"

#include "tet/descriptors/hogdescriptor.h"
#include "tet/encodings/kcencoding.h"

#include "opencv2/opencv.hpp"
#include "opencv2/highgui/highgui.hpp"

#include "tet/containers/view.h"
#include "tet/app/classifierapp.h"

#define BLOCK_DISP_SIZE 8
#define GAMMA 1

template <class _VecType>
float KCDistance(_VecType x, _VecType m)
{
    return std::exp(-(GAMMA / 2.0f)*(x - m).squaredNorm());
}

int main(int argc, char* argv[])
{
    if (argc < 5)
    {
        std::cout << "Usage: " << argv[0] << " [settings] [clusterID] [imfile] [visfile]" << std::endl;
        return -1;
    }

    typedef tet::classification::descriptors::HOGDescriptor DescriptorType;
    typedef tet::classification::encodings::KCEncoding EncoderType;
    
    int clusterID = boost::lexical_cast<int>(argv[2]);

    tet::classification::app::ClassifierApp<DescriptorType, EncoderType> app;
    app.readFile(argv[1]);

    tet::classification::containers::View<DescriptorType, EncoderType> view;
    view.setDescriptor(app.getDescriptor());
    view.setFilename(std::string(argv[3]));

    app.getDescriptor()->setRemoveZeroGradient(false);

    cv::Mat hogVis;
    DescriptorType::FeatureType descriptors = app.getDescriptor()->getDescriptors(*view.getInput(), &hogVis);
    EncoderType::FeatureType encoding = app.getEncoding()->computeFeatures(descriptors);
    //std::cout << encoding << std::endl;

    VlHog* hog = app.getDescriptor()->getHog();
    vl_size hogWidth = vl_hog_get_width(hog);
    vl_size hogHeight = vl_hog_get_height(hog);
    int hogNum = hogWidth * hogHeight;

    cv::Mat vis(hogHeight * BLOCK_DISP_SIZE, hogWidth * BLOCK_DISP_SIZE, CV_8UC3, cv::Scalar(0, 0, 0));

    float score = 0;
    for (size_t i = 0; i < hogHeight; ++i)
    {
        for (size_t j = 0; j < hogWidth; ++j)
        {
            if (clusterID >= 0)
            {
                std::vector<size_t> nn = app.getEncoding()->getTree()->findNN(descriptors.row(i * hogWidth + j), 1);

                if (nn[0] == clusterID)
                {
                    vis(cv::Rect(j*BLOCK_DISP_SIZE, i*BLOCK_DISP_SIZE, BLOCK_DISP_SIZE, BLOCK_DISP_SIZE)) = cv::Scalar(255, 255, 255);
                } else {
                    vis(cv::Rect(j*BLOCK_DISP_SIZE, i*BLOCK_DISP_SIZE, BLOCK_DISP_SIZE, BLOCK_DISP_SIZE)) = cv::Scalar(0, 0, 0);
                }
            } else {
                if (descriptors.row(i * hogWidth + j).block(0, 0, 1, descriptors.row(i * hogWidth + j).cols() - 2).norm() < 0.01)
                {
                    vis(cv::Rect(j*BLOCK_DISP_SIZE, i*BLOCK_DISP_SIZE, BLOCK_DISP_SIZE, BLOCK_DISP_SIZE)) = cv::Scalar(255, 0, 0);
                    continue;
                }

                std::vector<size_t> nn = app.getEncoding()->getTree()->findNN(descriptors.row(i * hogWidth + j), 4);
                //std::vector<std::vector<size_t> > nn = app.getEncoding()->getTree()->findNN(descriptors.row(i * hogWidth + j), 4);

                /*float max = 0;
                for (size_t i = 0; i < app.getEncoding()->getCenters().rows(); ++i)
                {
                    float dist = (descriptors.row(i * hogWidth + j) - app.getEncoding()->getCenters().row(i)).squaredNorm();
                    max = (dist > max) ? dist : max;
                }
                std::cout << max << std::endl;*/

                float weight = 0;
                for (size_t i = 0; i < nn.size(); ++i)
                {
                    float dist = KCDistance<EncoderType::SampleType>(descriptors.row(i * hogWidth + j),
                                                                     app.getEncoding()->getCenters().row(nn[i]));

                    //std::cout << dist << std::endl;
                    std::cout << dist << std::endl;
                    std::cout << app.getEncoding()->getFeatureWeights().size() << std::endl;
                    weight += app.getEncoding()->getFeatureWeights()[nn[i]]
                            * dist;
                }
                //std::cout << std::endl;
                
                //float weight = app.getEncoding()->getFeatureWeights()[encIdx];
                if (weight > 0)
                {
                    int colour = weight * 10 * 255;
                    vis(cv::Rect(j*BLOCK_DISP_SIZE, i*BLOCK_DISP_SIZE, BLOCK_DISP_SIZE, BLOCK_DISP_SIZE)) = cv::Scalar(0, colour, 0);
                } else {
                    int colour = -weight * 10 * 255;
                    vis(cv::Rect(j*BLOCK_DISP_SIZE, i*BLOCK_DISP_SIZE, BLOCK_DISP_SIZE, BLOCK_DISP_SIZE)) = cv::Scalar(0, 0, colour);
                }
            }
        }
    }

    /*for (size_t i = 0; i < hogHeight; ++i)
    {
        for (size_t j = 0; j < hogWidth; ++j)
        {
            std::stringstream norm;
            norm << std::setprecision(1) << descriptors.row(i * hogWidth + j).norm();
            std::cout << std::setprecision(1) << descriptors.row(i * hogWidth + j) << ":: " << descriptors.row(i * hogWidth + j).norm() << std::endl;
            cv::putText(vis, norm.str(), cv::Point(j * BLOCK_DISP_SIZE, i*BLOCK_DISP_SIZE), cv::FONT_HERSHEY_COMPLEX_SMALL, 0.5, cv::Scalar(255, 255, 255), 1);
        }
    }*/

    cv::imwrite(argv[4], vis);

    /*cv::namedWindow("Display window", CV_WINDOW_AUTOSIZE);
    cv::imshow("Display window", vis);

    cv::waitKey(0);*/
    
    return 0;

}
